# validated: 2017-12-09 EN f9bece2ffbf7 edu/wpi/first/wpilibj/livewindow/LiveWindow.java
# ----------------------------------------------------------------------------
#  Copyright (c) 2008-2017 FIRST. All Rights Reserved.                        
#  Open Source Software - may be modified and shared by FRC teams. The code   
#  must be accompanied by the FIRST BSD license file in the root directory of 
#  the project.                                                               
# ----------------------------------------------------------------------------
import threading
from networktables import NetworkTablesInstance
from .sendablebuilder import SendableBuilder

import warnings
import logging
logger = logging.getLogger(__name__)

__all__ = ["LiveWindow"]


class Component:
    def __init__(self, sendable, parent):
        self.sendable = sendable
        self.parent = parent
        self.builder = SendableBuilder()
        self.firstTime = True
        self.telemetryEnabled = True


class _LiveWindowComponent:
    """A LiveWindow component is a device (sensor or actuator) that should be
    added to the SmartDashboard in test mode. The components are cached until
    the first time the robot enters Test mode. This allows the components to
    be inserted, then renamed."""
    def __init__(self, subsystem, name, isSensor):
        self.subsystem = subsystem
        self.name = str(name)
        self.isSensor = isSensor

class LiveWindow:
    """The public interface for putting sensors and
    actuators on the LiveWindow."""

    components = {}
    _liveWindowTable = None
    _statusTable = None
    _enabledEntry = None
    startLiveWindow = False
    liveWindowEnabled = False
    telemetryEnabled = True
    mutex = threading.RLock()

    @classmethod
    def liveWindowTable(cls):
        if cls._liveWindowTable is None:
            cls._liveWindowTable = NetworkTablesInstance.getDefault().getTable("LiveWindow")
        return cls._liveWindowTable

    @classmethod
    def statusTable(cls):
        if cls._statusTable is None: 
            cls._statusTable = cls.liveWindowTable().getSubTable(".status")
        return cls._statusTable

    @classmethod
    def enabledEntry(cls):
        if cls._enabledEntry is None:
            cls._enabledEntry = cls.statusTable().getEntry("LW Enabled")
        return cls._enabledEntry
    
    @classmethod
    def _reset(cls):
        cls.components = {}
        cls._liveWindowTable = None
        cls._statusTable = None
        cls._enabledEntry = None
        cls.startLiveWindow = False
        cls.liveWindowEnabled = False
        cls.telemetryEnabled = True

    @classmethod
    def isEnabled(cls):
        return cls.liveWindowEnabled

    @classmethod
    def setEnabled(cls, enabled):
        """Set the enabled state of LiveWindow. If it's being enabled, turn
        off the scheduler and remove all the commands from the queue and
        enable all the components registered for LiveWindow. If it's being
        disabled, stop all the registered components and reenable the
        scheduler.

        TODO: add code to disable PID loops when enabling LiveWindow. The
        commands should reenable the PID loops themselves when they get
        rescheduled. This prevents arms from starting to move around, etc.
        after a period of adjusting them in LiveWindow mode.
        """
        from .command import Scheduler
        if cls.liveWindowEnabled != enabled:
            scheduler = Scheduler.getInstance()
            if enabled:
                logger.info("Starting live window mode.")
                scheduler.disable()
                scheduler.removeAll()
            else:
                logger.info("Stopping live window mode.")
                for component in cls.components.values():
                    component.builder.stopLiveWindowMode()
                scheduler.enable()
            cls.startLiveWindow = enabled
            cls.liveWindowEnabled = enabled
            cls.enabledEntry().setBoolean(enabled)

    @classmethod
    def run(cls):
        """The run method is called repeatedly to keep the values refreshed
        on the screen in test mode.

        .. deprecated:: 2018.0.0
            No longer required
        """
        warnings.warn("run is deprecated. It is no longer required.",
                      DeprecationWarning, stacklevel=2)
        cls.updateValues()

    @classmethod
    def addSensor(cls, subsystem, name, component):
        """Add a Sensor associated with the subsystem and with call it by the
        given name.

        :param subsystem: The subsystem this component is part of.
        :param name: The name of this component.
        :param component: A LiveWindowSendable component that represents a
            sensor.

        .. deprecated:: 2018.0.0
            Use :meth:`.Sendable.setName` instead.
        """
        warnings.warn("addSensor is deprecated. " + 
                      "Use Sendable.setName(subsystem, name) instead.",
                      DeprecationWarning, stacklevel=2)
        with cls.mutex:
            cls.add(component)
            component.setName(subsystem, name)

    @classmethod
    def addActuator(cls, subsystem, name, component):
        """Add an Actuator associated with the subsystem and with call it by
        the given name.

        :param subsystem: The subsystem this component is part of.
        :param name: The name of this component.
        :param component: A LiveWindowSendable component that represents a actuator.

        .. deprecated:: 2018.0.0
            Use :meth:`.Sendable.setName` instead.
        """
        warnings.warn("addActuator is deprecated. " +
                      "Use Sendable.setName(subsystem, name) instead.",
                      DeprecationWarning, stacklevel=2)
        with cls.mutex:
            cls.add(component)
            component.setName(subsystem, name)

    @classmethod
    def addSensorChannel(cls, moduleType, channel, component):
        """Add Sensor to LiveWindow. The components are shown with the type
        and channel like this: Gyro[0] for a gyro object connected to the
        first analog channel.

        :param moduleType: A string indicating the type of the module used in
            the naming (above)
        :param channel: The channel number the device is connected to
        :param component: A reference to the object being added

        .. deprecated:: 2018.0.0
            Use :meth:`.SendableBase.setName` instead.
        """
        warnings.warn("addSensorChannel is deprecated. "+ 
                      "Use SendableBase.setName(moduleType, channel) instead.",
                      DeprecationWarning, stacklevel=2)
        cls.add(component)
        component.setName("Ungrouped", "%s[%s]" % (moduleType, channel))

    @classmethod
    def addActuatorChannel(cls, moduleType, channel, component):
        """Add Actuator to LiveWindow. The components are shown with the
        module type, slot and channel like this: Servo[0,2] for a servo
        object connected to the first digital module and PWM port 2.

        :param moduleType: A string that defines the module name in the label
            for the value
        :param channel: The channel number the device is plugged into
            (usually PWM)
        :param component: The reference to the object being added

        .. deprecated:: 2018.0.0
            Use :meth:`.SendableBase.setName` instead.
        """
        warnings.warn("addActuatorChannel is deprecated. " + 
                      "Use SendableBase.setName(moduleType, channel) instead.",
                      DeprecationWarning, stacklevel=2)
        cls.add(component)
        component.setName("Ungrouped", "%s[%s]" % (moduleType, channel))

    @classmethod
    def addActuatorModuleChannel(cls, moduleType, moduleNumber, channel, component):
        """Add Actuator to LiveWindow. The components are shown with the
        module type, slot and channel like this: Servo[0,2] for a servo
        object connected to the first digital module and PWM port 2.

        :param moduleType: A string that defines the module name in the label
            for the value
        :param moduleNumber: The number of the particular module type
        :param channel: The channel number the device is plugged into
            (usually PWM)
        :param component: The reference to the object being added

        .. deprecated:: 2018.0.0
            Use :meth:`.SendableBase.setName` instead.
        """
        warnings.warn("addActuatorModuleChannel is deprecated. " + 
                      "Use SendableBase.setName(moduleType, moduleNumber, channel) instead.",
                      DeprecationWarning, stacklevel=2)
        cls.add(component)
        component.setName("Ungrouped", "%s[%s,%s]" % (moduleType, moduleNumber, channel))

    @classmethod
    def add(cls, sendable):
        """
        Add a component to the LiveWindow.

        :param sendable: component to add
        """
        with cls.mutex:
            if sendable not in cls.components:
                cls.components[sendable] = Component(sendable, None)

    @classmethod
    def addChild(cls, parent, child):
        """
        Add a child component to a component.

        :param parent: parent component
        :param child: child component
        """
        with cls.mutex:
            component = cls.components.get(child, None)
            if component is None:
                component = Component(None, parent)
                cls.components[child] = component
            else:
                component.parent = parent
            component.telemetryEnabled = False
        
    @classmethod
    def remove(cls, sendable):
        """
        Remove a component from the LiveWindow.

        @param sendable component to remove
        """
        with cls.mutex:
            if sendable in cls.components:
                component = cls.components[sendable]
                del cls.components[sendable]
                if cls.isEnabled():
                    component.builder.stopLiveWindowMode()

    @classmethod
    def enableTelemetry(cls, sendable):
        """
        Enable telemetry for a single component.

        :param sendable: component
        """
        with cls.mutex:
            cls.telemetryEnabled = True
            component = cls.components.get(sendable, None)
            if component is not None:
                component.telemetryEnabled = True

    @classmethod
    def disableTelemetry(cls, sendable):
        """
        Disable telemetry for a single component.  

        :param sendable: component
        """
        with cls.mutex:
            component = cls.components.get(sendable, None)
            if component is not None:
                component.telemetryEnabled = False

    @classmethod
    def disableAllTelemetry(cls):
        """ Disable ALL telemetry """
        with cls.mutex:
            cls.telemetryEnabled = False
            for component in cls.components.values():
                component.telemetryEnabled = False

    @classmethod
    def updateValues(cls):
        with cls.mutex:
            # only do this if either LiveWindow mode or telemetry is enabled
            if not cls.liveWindowEnabled and not cls.telemetryEnabled:
                return

            for component in cls.components.values():
                if component.sendable is not None and component.parent is None and (cls.liveWindowEnabled or component.telemetryEnabled):
                    if component.firstTime:
                        # By holding off creating the NetworkTable entries, it allows the
                        # components to be redefined. This allows default sensor and actuator
                        # values to be created that are replaced with the custom names from
                        # users calling setName.
                        name = component.sendable.getName()
                        if name == "":
                            continue
                        subsystem = component.sendable.getSubsystem()
                        ssTable = cls.liveWindowTable().getSubTable(subsystem)
                        if name == subsystem:
                            table = ssTable
                        else:
                            table = ssTable.getSubTable(name)
                        table.getEntry(".name").setString(name)
                        component.builder.setTable(table)
                        component.sendable.initSendable(component.builder)
                        ssTable.getEntry(".type").setString("LW Subsystem")

                        component.firstTime = False

                    if cls.startLiveWindow:
                        component.builder.startLiveWindowMode()
                    component.builder.updateTable()

            cls.startLiveWindow = False




# validated: 2017-10-03 EN e1195e8b9dab edu/wpi/first/wpilibj/command/TimedCommand.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .command import Command

__all__ = ["TimedCommand"]

class TimedCommand(Command):
    '''A command that runs for a set period of time.'''

    def __init__(self, name, timeoutInSeconds):
        """Instantiates a TimedCommand with the given name and timeout.

        :param name: the name of the command
        :param timeoutInSeconds: the time the command takes to run
        """
        super().__init__(name, timeoutInSeconds)


    def isFinished(self):
        """Ends command when timed out."""
        return self.isTimedOut()



# validated: 2017-10-19 AA 4e80570c4c48 edu/wpi/first/wpilibj/buttons/NetworkButton.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .button import Button

from networktables.networktable import NetworkTable
from networktables import NetworkTablesInstance

__all__ = ["NetworkButton"]


class NetworkButton(Button):
    """A :class:`.button.Button` that uses a :class:`.NetworkTable` boolean field."""

    def __init__(self, table, field):
        """Initialize the NetworkButton.

        :param table: the :class:`.NetworkTable` instance to use, or the name of the
                      table to use.
        :param field: field to use.
        """

        if isinstance(table, NetworkTable):
            self.entry = table.getEntry(field)
        else:
            table = NetworkTablesInstance.getDefault().getTable(table)
            self.entry = table.getEntry(field)

    def get(self):
        """Get the value of the button."""
        return (self.entry.getInstance().isConnected()
                and self.entry.getBoolean(False))



# validated: 2017-11-09 TW ef3267833fc3 edu/wpi/first/wpilibj/TimedRobot.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal

from .iterativerobotbase import IterativeRobotBase
from .notifier import Notifier
from .timer import Timer


__all__ = ["TimedRobot"]


class TimedRobot(IterativeRobotBase):
    """TimedRobot implements the IterativeRobotBase robot program framework.

    The TimedRobot class is intended to be subclassed by a user creating a robot program.

    periodic() functions from the base class are called on an interval by a Notifier instance.
    """
    DEFAULT_PERIOD = .02


    def __init__(self):
        super().__init__()
        hal.report(hal.UsageReporting.kResourceType_Framework, hal.UsageReporting.kFramework_Iterative)

        self.period = TimedRobot.DEFAULT_PERIOD
        # Prevents loop from starting if user calls setPeriod() in robotInit()
        self.startLoop = False
        self.loop = Notifier(self.loopFunc)

    def startCompetition(self) -> None:
        """Provide an alternate "main loop" via startCompetition()"""
        self.robotInit()

        hal.observeUserProgramStarting()

        self.startLoop = True
        self.loop.startPeriodic(self.period)

        while True:
            Timer.delay(60*60*24)


    def setPeriod(self, period: float) -> None:
        """Set time period between calls to Periodic() functions.

        :param period: Period in seconds
        """
        self.period = period

        if self.startLoop:
            self.loop.startPeriodic(self.period)



# validated: 2018-01-19 DS 0e8ff4663df9 edu/wpi/first/wpilibj/SpeedControllerGroup.java

# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from .interfaces.speedcontroller import SpeedController
from .sendablebase import SendableBase

__all__ = ["SpeedControllerGroup"]


class SpeedControllerGroup(SendableBase, SpeedController):
    """Allows multiple :class:`.SpeedController` objects to be linked together."""

    instances = 0

    def __init__(self, speedController, *args):
        """Create a new SpeedControllerGroup with the provided SpeedControllers.

        :param args: SpeedControllers to add
        :type args: :class:`.SpeedController`
        """
        SendableBase.__init__(self)
        SpeedController.__init__(self)

        self.speedControllers = [speedController] + list(args)

        for speedcontroller in self.speedControllers:
            self.addChild(speedcontroller)

        self.isInverted = False

        SpeedControllerGroup.instances += 1
        self.setName("SpeedControllerGroup", self.instances)

    def set(self, speed):
        for speedController in self.speedControllers:
            speedController.set(-speed if self.isInverted else speed)

    def get(self):
        if len(self.speedControllers) > 0:
            return self.speedControllers[0].get() * (-1 if self.isInverted else 1)
        return 0.0

    def setInverted(self, isInverted):
        self.isInverted = isInverted

    def getInverted(self):
        return self.isInverted

    def stopMotor(self):
        for speedController in self.speedControllers:
            speedController.stopMotor()

    def disable(self):
        for speedController in self.speedControllers:
            speedController.disable()

    def pidWrite(self, output):
        self.set(output)

    def initSendable(self, builder):
        builder.setSmartDashboardType("Speed Controller")
        builder.setSafeState(self.stopMotor)
        builder.addDoubleProperty("Value", self.get, self.set)



# validated: 2017-09-27 AA e1195e8b9dab edu/wpi/first/wpilibj/interfaces/Accelerometer.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in $(WIND_BASE)/WPILib.
#----------------------------------------------------------------------------

__all__ = ["Accelerometer"]

class Accelerometer:
    """Interface for 3-axis accelerometers"""

    class Range:
        k2G = 0
        k4G = 1
        k8G = 2
        k16G = 3

    def setRange(self, range):
        """Common interface for setting the measuring range of an
        accelerometer.

        :param range: The maximum acceleration, positive or negative, that
                      the accelerometer will measure.  Not all accelerometers
                      support all ranges.
        """
        raise NotImplementedError

    def getX(self):
        """Common interface for getting the x axis acceleration

        :returns: The acceleration along the x axis in g-forces
        """
        raise NotImplementedError

    def getY(self):
        """Common interface for getting the y axis acceleration

        :returns: The acceleration along the y axis in g-forces
        """
        raise NotImplementedError

    def getZ(self):
        """Common interface for getting the z axis acceleration

        :returns: The acceleration along the z axis in g-forces
        """
        raise NotImplementedError



# validated: 2017-12-08 DV 85157a56c3a7 edu/wpi/first/wpilibj/filters/LinearDigitalFilter.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import collections
import math
from operator import mul

from .filter import Filter

__all__ = ['LinearDigitalFilter']

class LinearDigitalFilter(Filter):
    """This class implements a linear, digital filter. All types of FIR and IIR
    filters are supported. Static factory methods are provided to create commonly
    used types of filters.
    
    Filters are of the form::
    
        y[n] = (b0*x[n] + b1*x[n-1] + ... + bP*x[n-P]) - (a0*y[n-1] + a2*y[n-2] + ... + aQ*y[n-Q])
    
    Where:
    
    * ``y[n]`` is the output at time "n"
    * ``x[n]`` is the input at time "n"
    * ``y[n-1]`` is the output from the LAST time step ("n-1")
    * ``x[n-1]`` is the input from the LAST time step ("n-1")
    * ``b0...bP`` are the "feedforward" (FIR) gains
    * ``a0...aQ`` are the "feedback" (IIR) gains
    
    .. note:: IMPORTANT! Note the "-" sign in front of the feedback term! This is a common
              convention in signal processing.
    
    What can linear filters do? Basically, they can filter, or diminish, the
    effects of undesirable input frequencies. High frequencies, or rapid changes,
    can be indicative of sensor noise or be otherwise undesirable. A "low pass"
    filter smoothes out the signal, reducing the impact of these high frequency
    components.  Likewise, a "high pass" filter gets rid of slow-moving signal
    components, letting you detect large changes more easily.
    
    Example FRC applications of filters:
    
    * Getting rid of noise from an analog sensor input (note: the roboRIO's FPGA
      can do this faster in hardware)
    * Smoothing out joystick input to prevent the wheels from slipping or the
      robot from tipping
    * Smoothing motor commands so that unnecessary strain isn't put on
      electrical or mechanical components
    * If you use clever gains, you can make a PID controller out of this class!
    
    For more on filters, I highly recommend the following articles:
    
    * http://en.wikipedia.org/wiki/Linear_filter
    * http://en.wikipedia.org/wiki/Iir_filter
    * http://en.wikipedia.org/wiki/Fir_filter
    
    .. note:: :meth:`pidGet` should be called by the user on a known, regular period.
        You can set up a Notifier to do this (look at the :class:`.PIDController`
        class), or do it "inline" with code in a periodic function.
    
    .. note:: For ALL filters, gains are necessarily a function of frequency. If
        you make a filter that works well for you at, say, 100Hz, you will most
        definitely need to adjust the gains if you then want to run it at 200Hz!
        Combining this with Note 1 - the impetus is on YOU as a developer to make
        sure :meth:`pidGet` gets called at the desired, constant frequency!
        
    There are static methods you can use to build common filters:
    
    * :meth:`highPass`
    * :meth:`movingAverage`
    * :meth:`singlePoleIIR`
    
    """
    
    def __init__(self, source, ffGains, fbGains):
        """Constructor. Create a linear FIR or IIR filter
        
        :param source: The PIDSource object that is used to get values
        :type source: :class:`.PIDSource`, callable
        :param ffGains: The "feed forward" or FIR gains
        :type ffGains: list, tuple
        :param fbGains: The "feed back" or IIR gains
        :type fbGains: list, tuple
        """
        
        super().__init__(source)
    
        lf = len(ffGains)
        lb = len(fbGains)
    
        self.inputs = collections.deque([0.0]*lf, maxlen=lf)
        self.outputs = collections.deque([0.0]*lb, maxlen=lb)
        self.inputGains = ffGains
        self.outputGains = fbGains
    
    @staticmethod
    def singlePoleIIR(source, timeConstant, period):
        """Creates a one-pole IIR low-pass filter of the form::
        
            y[n] = (1-gain)*x[n] + gain*y[n-1]
            
        Where ``gain = e^(-dt / T)``, ``T`` is the time constant in seconds
        
        This filter is stable for time constants greater than zero
        
        :param source: The PIDSource object that is used to get values
        :type source: :class:`.PIDSource`, callable
        :param timeConstant: The discrete-time time constant in seconds
        :type timeConstant: float
        :param period: The period in seconds between samples taken by the user
        :type period: float
        
        :returns: :class:`LinearDigitalFilter`
        """
        
        gain = math.exp(-period/float(timeConstant))
        ffGains = [1.0 - gain]
        fbGains = [-gain]
        
        return LinearDigitalFilter(source, ffGains, fbGains)

    @staticmethod
    def highPass(source, timeConstant, period):
        """Creates a first-order high-pass filter of the form::
        
            y[n] = gain*x[n] + (-gain)*x[n-1] + gain*y[n-1]
            
        where ``gain = e^(-dt / T)``, ``T`` is the time constant in seconds
        
        This filter is stable for time constants greater than zero
        
        :param source: The PIDSource object that is used to get values
        :type source: :class:`.PIDSource`, callable
        :param timeConstant: The discrete-time time constant in seconds
        :type timeConstant: float
        :param period: The period in seconds between samples taken by the user
        :type period: float
        
        :returns: :class:`LinearDigitalFilter`
        """
        gain = math.exp(-period/float(timeConstant))
        ffGains = [gain, -gain]
        fbGains = [-gain]
            
        return LinearDigitalFilter(source, ffGains, fbGains)
    
    @staticmethod
    def movingAverage(source, taps):
        """Creates a K-tap FIR moving average filter of the form::
        
            y[n] = 1/k * (x[k] + x[k-1] + ... + x[0])
        
        This filter is always stable.
        
        :param source: The PIDSource object that is used to get values
        :type source: :class:`.PIDSource`, callable
        :param taps: The number of samples to average over. Higher = smoother but slower
        :type taps: int
        
        :raises: :exc:`ValueError` if number of taps is less than 1
        
        :returns: :class:`LinearDigitalFilter`
        """
        if taps <= 0:
            raise ValueError("Number of taps was not at least 1")
        
        ffGains = [1.0/taps]*taps
        fbGains = []

        return LinearDigitalFilter(source, ffGains, fbGains)
    
    def get(self):
        """Returns the current filter estimate without also inserting new data as
        :meth:`pidGet` would do.
        
        :returns: The current filter estimate
        """

        # Calculate the new value
        retVal = sum(map(mul, self.inputs, self.inputGains)) - \
                 sum(map(mul, self.outputs, self.outputGains)) 
        
        return retVal

    def reset(self):
        """Reset the filter state"""
        self.inputs.clear()
        self.outputs.clear()
     
    def pidGet(self):
        """Calculates the next value of the filter
        
        :returns: The filtered value at this step
        """
        
        # Rotate the inputs
        self.inputs.appendleft(self.pidGetSource())
    
        # Calculate the new value
        retVal = sum(map(mul, self.inputs, self.inputGains)) - \
                 sum(map(mul, self.outputs, self.outputGains))
    
        # Rotate the outputs
        self.outputs.appendleft(retVal)
    
        return retVal



# validated: 2017-12-21 DV de134a5c608d edu/wpi/first/wpilibj/livewindow/LiveWindowSendable.java
#----------------------------------------------------------------------------
# Copyright (c) 2008-2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import warnings

from networktables import NetworkTables
from .sendable import Sendable

__all__ = ["LiveWindowSendable"]

class LiveWindowSendable(Sendable):
    """A special type of object that can be displayed on the live window.

    .. deprecated:: 2018.0
        Use :class:`.Sendable` directly instead.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        warnings.warn('LiveWindowSendable is deprecated, use Sendable directly instead',
                      DeprecationWarning, stacklevel=2)

    def updateTable(self):
        """Update the table for this sendable object with the latest
        values.
        """
        raise NotImplementedError

    def startLiveWindowMode(self):
        """Start having this sendable object automatically respond to
        value changes reflect the value on the table.

        Default implementation will add self.valueChanged (if it exists)
        as a table listener on "Value".
        """
        if hasattr(self, "valueChanged"):
            valueEntry = getattr(self, "valueEntry", None)
            valueListener = getattr(self, "valueListener", None)
            if valueEntry is None or valueListener is not None:
                return
            self.valueListener = valueEntry.addListener(
                self.valueChanged, 
                NetworkTables.NotifyFlags.IMMEDIATE |
                NetworkTables.NotifyFlags.NEW |
                NetworkTables.NotifyFlags.UPDATE)

    def stopLiveWindowMode(self):
        """Stop having this sendable object automatically respond to value
        changes.
        """
        # TODO: Broken, should only remove the listener from "Value" only.
        valueEntry = getattr(self, "valueEntry", None)
        valueListener = getattr(self, "valueListener", None)
        if valueEntry is None or valueListener is None:
            return
        valueEntry.removeListener(valueListener)
        self.valueListener = None

    def getName(self):
        return ""

    def _setName(self, name):
        pass

    def getSubsystem(self):
        return ""

    def setSubsystem(self, subsystem):
        pass

    def initSendable(self, builder):
        builder.setUpdateTable(self.updateTable)



# validated: 2017-12-12 EN f9bece2ffbf7 edu/wpi/first/wpilibj/SolenoidBase.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .sendablebase import SendableBase

__all__ = ["SolenoidBase"]

class SolenoidBase(SendableBase):
    """SolenoidBase class is the common base class for the Solenoid and
    DoubleSolenoid classes."""

    def __init__(self, moduleNumber):
        """Constructor.

        :param moduleNumber: The PCM CAN ID
        """
        super().__init__()
        self.moduleNumber = moduleNumber

    def getAll(moduleNumber):
        """Read all 8 solenoids from the specified module as a
        single byte.

        :param moduleNumber: in a static context, the module number to read. otherwise don't provide it.
        :returns: The current value of all 8 solenoids on the module.
        """
        if isinstance(moduleNumber, SolenoidBase):
            moduleNumber = moduleNumber.moduleNumber
        return hal.getAllSolenoids(moduleNumber)

    def getPCMSolenoidBlackList(moduleNumber):
        """
        Reads complete solenoid blacklist for all 8 solenoids as a single byte.
            If a solenoid is shorted, it is added to the blacklist and
            disabled until power cycle, or until faults are cleared. See
            :meth:`clearAllPCMStickyFaults`

        :param moduleNumber: in a static context, the module number to read. otherwise don't provide it.
        :returns: The solenoid blacklist of all 8 solenoids on the module.
        """
        if isinstance(moduleNumber, SolenoidBase):
            moduleNumber = moduleNumber.moduleNumber
        return hal.getPCMSolenoidBlackList(moduleNumber)

    def getPCMSolenoidVoltageStickyFault(moduleNumber):
        """
        :param moduleNumber: in a static context, the module number to read. otherwise don't provide it.
        :returns: True if PCM Sticky fault is set : The common
            highside solenoid voltage rail is too low, most likely
            a solenoid channel has been shorted.
        """
        if isinstance(moduleNumber, SolenoidBase):
            moduleNumber = moduleNumber.moduleNumber
        return hal.getPCMSolenoidVoltageStickyFault(moduleNumber)

    def getPCMSolenoidVoltageFault(moduleNumber):
        """
        :param moduleNumber: in a static context, the module number to read. otherwise don't provide it.
        :returns: True if PCM is in fault state : The common
            highside solenoid voltage rail is too low, most likely
            a solenoid channel has been shorted.
        """
        if isinstance(moduleNumber, SolenoidBase):
            moduleNumber = moduleNumber.moduleNumber
        return hal.getPCMSolenoidVoltageFault(moduleNumber)

    def clearAllPCMStickyFaults(moduleNumber):
        """
        Clear ALL sticky faults inside the PCM that Solenoid is wired to.

        If a sticky fault is set, then it will be persistently cleared. Compressor drive
            maybe momentarily disable while flages are being cleared. Care should be
            taken to not call this too frequently, otherwise normal compressor functionality
            may be prevented.

        If no sticky faults are set then this call will have no effect.

        :param moduleNumber: in a static context, the module number to read. otherwise don't provide it.
        """
        if isinstance(moduleNumber, SolenoidBase):
            moduleNumber = moduleNumber.moduleNumber
        hal.clearAllPCMStickyFaults(moduleNumber)



# validated: 2017-12-25 TW f9bece2ffbf7 edu/wpi/first/wpilibj/drive/KilloughDrive.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import math

from .robotdrivebase import RobotDriveBase
from .vector2d import Vector2d

__all__ = ["KilloughDrive"]


class KilloughDrive(RobotDriveBase):
    r"""A class for driving Killough drive platforms.

    Killough drives are triangular with one omni wheel on each corner.

    Drive Base Diagram::

          /_____\
         / \   / \
            \ /
            ---

    Each `drive()` function provides different inverse kinematic relations for a Killough drive.
    The default wheel vectors are parallel to their respective opposite sides, but can be overridden.
    See the constructor for more information.

    This library uses the NED axes convention (North-East-Down as external reference in the world
    frame): http://www.nuclearprojects.com/ins/images/axis_big.png.

    The positive X axis points ahead, the positive Y axis points right, and the positive Z axis
    points down. Rotations follow the right-hand rule, so clockwise rotation around the Z axis is
    positive.
    """

    kDefaultLeftMotorAngle = 60.0
    kDefaultRightMotorAngle = 120.0
    kDefaultBackMotorAngle = 270.0

    instances = 0

    def __init__(self, leftMotor, rightMotor, backMotor, leftMotorAngle=kDefaultLeftMotorAngle,
                 rightMotorAngle=kDefaultRightMotorAngle, backMotorAngle=kDefaultBackMotorAngle):
        """Construct a Killough drive with the given motors and default motor angles.

        Angles are measured in degrees clockwise from the positive X axis.
        
        The default motor angles make the wheels on each corner parallel to their
        respective opposite sides.

        If a motor needs to be inverted, do so before passing it in.

        :param leftMotor: The motor on the left corner.
        :param rightMotor: The motor on the right corner.
        :param backMotor: The motor on the back corner.
        :param leftMotorAngle: The angle of the left wheel's forward direction of travel
        :param rightMotorAngle: The angle of the right wheel's forward direction of travel
        :param backMotorAngle: The angle of the back wheel's forward direction of travel
        """
        super().__init__()

        self.leftMotor = leftMotor
        self.rightMotor = rightMotor
        self.backMotor = backMotor

        self.leftVec = Vector2d(math.cos(math.radians(leftMotorAngle)),
                                math.sin(math.radians(leftMotorAngle)))
        self.rightVec = Vector2d(math.cos(math.radians(rightMotorAngle)),
                                 math.sin(math.radians(rightMotorAngle)))
        self.backVec = Vector2d(math.cos(math.radians(backMotorAngle)),
                                math.sin(math.radians(backMotorAngle)))

        self.addChild(self.leftMotor)
        self.addChild(self.rightMotor)
        self.addChild(self.backMotor)
        KilloughDrive.instances += 1
        self.setName("KilloughDrive", self.instances)

        self.reported = False

    def driveCartesian(self, ySpeed, xSpeed, zRotation, gyroAngle=0.0):
        """Drive method for Killough platform.

        Angles are measured clockwise from the positive X axis. The robot's speed is independent
        from its angle or rotation rate.

        :param ySpeed: The robot's speed along the Y axis `[-1.0..1.0]`. Right is positive.
        :param xSpeed: The robot's speed along the X axis `[-1.0..1.0]`. Forward is positive.
        :param zRotation: The robot's rotation rate around the Z axis `[-1.0..1.0]`. Clockwise is positive.
        :param gyroAngle: The current angle reading from the gyro in degrees around the Z axis. Use
                          this to implement field-oriented controls.
        """

        if not self.reported:
            # hal.report(hal.UsageReporting.kResourceType_RobotDrive,
            #           3,
            #           hal.UsageReporting.kRobotDrive_Curvature)
            self.reported = True

        ySpeed = RobotDriveBase.limit(ySpeed)
        ySpeed = RobotDriveBase.applyDeadband(ySpeed, self.deadband)

        xSpeed = RobotDriveBase.limit(xSpeed)
        xSpeed = RobotDriveBase.applyDeadband(xSpeed, self.deadband)

        # Compensate for gyro angle
        input = Vector2d(ySpeed, xSpeed)
        input.rotate(gyroAngle)

        wheelSpeeds = [input.scalarProject(self.leftVec) + zRotation,
                       input.scalarProject(self.rightVec) + zRotation,
                       input.scalarProject(self.backVec) + zRotation]

        RobotDriveBase.normalize(wheelSpeeds)

        self.leftMotor.set(wheelSpeeds[0] * self.maxOutput)
        self.rightMotor.set(wheelSpeeds[1] * self.maxOutput)
        self.backMotor.set(wheelSpeeds[2] * self.maxOutput)

        self.feed()

    def drivePolar(self, magnitude, angle, zRotation):
        """Drive method for Killough platform.

        Angles are measured counter-clockwise from straight ahead. The speed at which the robot
        drives (translation) is independent from its angle or zRotation rate.

        :param magnitude: The robot's speed at a given angle `[-1.0..1.0]`. Forward is positive.
        :param angle: The angle around the Z axis at which the robot drives in degrees `[-180..180]`.
        :param zRotation: The robot's rotation rate around the Z axis `[-1.0..1.0]`. Clockwise is positive.
        """
        if not self.reported:
            # hal.report(hal.UsageReporting.kResourceType_RobotDrive,
            #           3,
            #           hal.UsageReporting.kRobotDrive_KilloughPolar)
            self.reported = True

        magnitude = RobotDriveBase.limit(magnitude) * math.sqrt(2)

        self.driveCartesian(magnitude * math.cos(math.radians(angle)), magnitude * math.sin(math.radians(angle)),
                            zRotation, 0)

    def stopMotor(self):
        self.leftMotor.stopMotor()
        self.rightMotor.stopMotor()
        self.backMotor.stopMotor()
        self.feed()

    def getDescription(self):
        return "Killough Drive"

    def initSendable(self, builder):
        builder.setSmartDashboardType("KilloughDrive")
        builder.addDoubleProperty("Left Motor Speed", self.leftMotor.get, self.leftMotor.set)
        builder.addDoubleProperty("Right Motor Speed", self.rightMotor.get, self.rightMotor.set)
        builder.addDoubleProperty("Back Motor Speed", self.backMotor.get, self.backMotor.set)



# validated: 2017-11-13 TW 21585f70a88e edu/wpi/first/wpilibj/JoystickBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------
import warnings

from .generichid import GenericHID

__all__ = ["JoystickBase"]

class JoystickBase(GenericHID):
    """
    JoystickBase Interface.

    ..deprecated:: 2018.0.0
        Inherit directly from GenericHID instead.
    """

    def __init__(self, port):
        warnings.warn("JoystickBase is deprecated. Inherit directly from GenericHID instead", DeprecationWarning, stacklevel=2)
        super().__init__(port)

    def getZ(self, hand=None):
        """Get the z position of the HID.

        :param hand: which hand, left or right
        :returns: the z position
        """

        raise NotImplementedError

    def getTwist(self):
        """Get the twist value.

        :returns: the twist value
        """

        raise NotImplementedError

    def getThrottle(self):
        """Get the throttle.

        :returns: the throttle value
        """

        raise NotImplementedError



# validated: 2018-01-06 TW f9bece2ffbf7 edu/wpi/first/wpilibj/Sendable.java

from .sendablebuilder import SendableBuilder

__all__ = ["Sendable"]


class Sendable:
    """The base interface for objects that can be sent over the network
    through network tables"""

    def getName(self) -> str:
        """
        Gets the name of this Sendable object.

        :returns: Name
        """
        raise NotImplementedError

    def setName(self, subsystem: str, name: str = None) -> None:
        """
        Sets the name (and optionally the subsystem name) of this Sendable object.

        This may be called with two different sets of parameters:

        - name
        - subsystem, name

        :param str subsystem: subsystem name
        :param str name: Name
        """
        if name is None:
            self._setName(subsystem)
        else:
            self._setNameAndSubsystem(subsystem, name)

    def _setName(self, name: str) -> None:
        """Sets the name of this Sendable object."""
        raise NotImplementedError

    def _setNameAndSubsystem(self, subsystem: str, name: str) -> None:
        """
        Sets both the subsystem name and device name of this Sendable object.

        :param subsystem: subsystem name
        :param name: Name
        """
        self.setSubsystem(subsystem)
        self.setName(name)

    def getSubsystem(self) -> str:
        """
        Gets the subsystem name of this Sendable object.

        :returns: subsystem name
        """
        raise NotImplementedError

    def setSubsystem(self, subsystem: str) -> None:
        """
        Sets the subsystem name of this Sendable object.

        :param subsystem: subsystem name
        """
        raise NotImplementedError

    def initSendable(self, builder: SendableBuilder) -> None:
        """
        Initializes this Sendable object.

        :param builder: sendable builder
        """
        raise NotImplementedError



# validated: 2017-12-12 EN f9bece2ffbf7 edu/wpi/first/wpilibj/PowerDistributionPanel.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from functools import partial
from .sensorbase import SensorBase

__all__ = ["PowerDistributionPanel"]

class PowerDistributionPanel(SensorBase):
    """
        Use to obtain voltage, current, temperature, power, and energy from the
        Power Distribution Panel over CAN
    """

    def __init__(self, module=0):
        """
            :param module: CAN ID of the PDP
            :type module: int
        """
        super().__init__()
        self.module = module
        SensorBase.checkPDPModule(module)
        hal.initializePDP(module)
        self.setName("PowerDistributionPanel", module)

    def getVoltage(self):
        """
            Query the input voltage of the PDP

            :returns: The voltage of the PDP in volts
            :rtype: float
        """
        return hal.getPDPVoltage(self.module)

    def getTemperature(self):
        """
            Query the temperature of the PDP

            :returns: The temperature of the PDP in degrees Celsius
            :rtype: float
        """
        return hal.getPDPTemperature(self.module)

    def getCurrent(self, channel):
        """
            Query the current of a single channel of the PDP

            :returns: The current of one of the PDP channels (channels 0-15)
                      in Amperes
            :rtype: float
        """
        SensorBase.checkPDPChannel(channel)
        return hal.getPDPChannelCurrent(self.module, channel)
    
    def getTotalCurrent(self):
        """
            Query the current of all monitored PDP channels (0-15)

            :returns: The total current drawn from the PDP channels in Amperes
            :rtype: float
        """
        return hal.getPDPTotalCurrent(self.module)
    
    def getTotalPower(self):
        """
            Query the total power drawn from the monitored PDP channels

            :returns: The total power drawn from the PDP channels in Watts
            :rtype: float
        """
        return hal.getPDPTotalPower(self.module)
    
    def getTotalEnergy(self):
        """
            Query the total energy drawn from the monitored PDP channels

            :returns: The total energy drawn from the PDP channels in Joules
            :rtype: float
        """
        return hal.getPDPTotalEnergy(self.module)
    
    def resetTotalEnergy(self):
        """
            Reset the total energy to 0
        """
        hal.resetPDPTotalEnergy(self.module)
    
    def clearStickyFaults(self):
        """
            Clear all pdp sticky faults
        """
        hal.clearPDPStickyFaults(self.module)

    def initSendable(self, builder):
        builder.setSmartDashboardType("PowerDistributionPanel")
        for chan in range(self.kPDPChannels):
            builder.addDoubleProperty("Chan%s" % (chan,), partial(self.getCurrent, chan), None)
        builder.addDoubleProperty("Voltage", self.getVoltage, None)
        builder.addDoubleProperty("TotalCurrent", self.getTotalCurrent, None)



# validated: 2017-12-27 TW f9bece2ffbf7 edu/wpi/first/wpilibj/AnalogAccelerometer.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .analoginput import AnalogInput
from .interfaces import PIDSource
from .sensorbase import SensorBase

__all__ = ["AnalogAccelerometer"]

class AnalogAccelerometer(SensorBase):
    """Analog Accelerometer
    
    The accelerometer reads acceleration directly through the sensor. Many
    sensors have multiple axis and can be treated as multiple devices. Each
    is calibrated by finding the center value over a period of time.
    
    .. not_implemented: initAccelerometer
    """
    
    PIDSourceType = PIDSource.PIDSourceType

    def __init__(self, channel):
        """Constructor. Create a new instance of Accelerometer from either an existing
        AnalogChannel or from an analog channel port index.

        :param channel: port index or an already initialized AnalogInput
        :type channel: int or :class:`.AnalogInput`
        """
        super().__init__()
        if not hasattr(channel, "getAverageVoltage"): # If 'channel' is an integer
            self.analogChannel = AnalogInput(channel)
            self.allocatedChannel = True
            self.addChild(self.analogChannel)
        else:
            self.allocatedChannel = False
            self.analogChannel = channel
        self.voltsPerG = 1.0
        self.zeroGVoltage = 2.5
        self.pidSource = self.PIDSourceType.kDisplacement
        hal.report(hal.UsageReporting.kResourceType_Accelerometer,
                      self.analogChannel.getChannel())
        self.setName("Accelerometer", self.analogChannel.getChannel())

    def free(self):
        super().free()
        if self.analogChannel and self.allocatedChannel:
            self.analogChannel.free()
        self.analogChannel = None


    def getAcceleration(self):
        """Return the acceleration in Gs.

        The acceleration is returned units of Gs.

        :returns: The current acceleration of the sensor in Gs.
        :rtype: float
        """
        if not self.analogChannel:
            return 0.0
        return (self.analogChannel.getAverageVoltage() - self.zeroGVoltage) / self.voltsPerG

    def setSensitivity(self, sensitivity):
        """Set the accelerometer sensitivity.

        This sets the sensitivity of the accelerometer used for calculating
        the acceleration.  The sensitivity varies by accelerometer model.
        There are constants defined for various models.

        :param sensitivity: The sensitivity of accelerometer in Volts per G.
        :type  sensitivity: float
        """
        self.voltsPerG = sensitivity

    def setZero(self, zero):
        """Set the voltage that corresponds to 0 G.

        The zero G voltage varies by accelerometer model. There are constants
        defined for various models.

        :param zero: The zero G voltage.
        :type  zero: float
        """
        self.zeroGVoltage = zero
        
    def setPIDSourceType(self, pidSource):
        """Set which parameter you are using as a process
        control variable. 

        :param pidSource: An enum to select the parameter.
        :type  pidSource: :class:`.PIDSource.PIDSourceType`
        """
        self.pidSource = pidSource
        
    def getPIDSourceType(self):
        return self.pidSource

    def pidGet(self):
        """Get the Acceleration for the PID Source parent.

        :returns: The current acceleration in Gs.
        :rtype: float
        """
        return self.getAcceleration()

    def initSendable(self, builder):
        builder.setSmartDashboardType("Accelerometer")
        builder.addDoubleProperty("Value", self.getAcceleration, None)



# validated: 2017-12-06 DV dd7563376bf6 edu/wpi/first/wpilibj/RobotBase.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
from networktables import NetworkTables

import logging
logger = logging.getLogger('robotpy')

__all__ = ["RobotBase"]

class RobotBase:
    """Implement a Robot Program framework.

    The RobotBase class is intended to be subclassed by a user creating a
    robot program.  Overridden ``autonomous()`` and ``operatorControl()`` methods
    are called at the appropriate time as the match proceeds. In the current
    implementation, the Autonomous code will run to completion before the
    OperatorControl code could start. In the future the Autonomous code might
    be spawned as a task, then killed at the end of the Autonomous period.

    User code should be placed in the constructor that runs before the
    Autonomous or Operator Control period starts. The constructor will
    run to completion before Autonomous is entered.

    .. warning:: If you override ``__init__`` in your robot class, you must call
                 the base class constructor. This must be used to ensure that
                 the communications code starts.

    .. not_implemented: getBooleanProperty
    """

    def __init__(self):
        NetworkTables.setNetworkIdentity("Robot")

        if self.isReal():
            NetworkTables.startServer("/home/lvuser/networktables.ini")
        else:
            NetworkTables.startServer()

        from .driverstation import DriverStation
        self.ds = DriverStation.getInstance()

        NetworkTables.getTable("LiveWindow").getSubTable(".status").getEntry("LW Enabled").setBoolean(False)
        from .livewindow import LiveWindow
        LiveWindow.setEnabled(False)

        self.__initialized = True

    def free(self):
        """Free the resources for a RobotBase class."""
        pass

    @staticmethod
    def isSimulation() -> bool:
        """Get if the robot is a simulation.

        :returns: If the robot is running in simulation.
        """
        return hal.isSimulation()

    @staticmethod
    def isReal() -> bool:
        """Get if the robot is real.

        :returns: If the robot is running in the real world.
        """
        return not hal.isSimulation()

    def isDisabled(self) -> bool:
        """Determine if the Robot is currently disabled.

        :returns: True if the Robot is currently disabled by the field controls.
        """
        return self.ds.isDisabled()

    def isEnabled(self) -> bool:
        """Determine if the Robot is currently enabled.

        :returns: True if the Robot is currently enabled by the field controls.
        """
        return self.ds.isEnabled()

    def isAutonomous(self) -> bool:
        """Determine if the robot is currently in Autonomous mode as
        determined by the field controls.

        :returns: True if the robot is currently operating Autonomously
        """
        return self.ds.isAutonomous()

    def isTest(self) -> bool:
        """Determine if the robot is currently in Test mode as
        determined by the driver station.

        :returns: True if the robot is currently operating in Test mode.
        """
        return self.ds.isTest()

    def isOperatorControl(self) -> bool:
        """Determine if the robot is currently in Operator Control mode as
        determined by the field controls.

        :returns: True if the robot is currently operating in Tele-Op mode
        """
        return self.ds.isOperatorControl()

    def isNewDataAvailable(self) -> bool:
        """Indicates if new data is available from the driver station.

        :returns: Has new data arrived over the network since the last time
                  this function was called?
        """
        return self.ds.isNewControlData()

    def startCompetition(self):
        """Provide an alternate "main loop" via startCompetition()."""
        raise NotImplementedError

    @staticmethod
    def initializeHardwareConfiguration():
        """Common initialization for all robot programs."""

        # Python specific: do not call this, initialize() is already called when
        # hal is imported
        #hal.initialize()

        from .driverstation import DriverStation
        from .robotstate import RobotState
        RobotState.impl = DriverStation.getInstance()

    @staticmethod
    def main(robot_cls):
        """Starting point for the applications."""
        RobotBase.initializeHardwareConfiguration()

        hal.report(hal.UsageReporting.kResourceType_Language,
                   hal.UsageReporting.kLanguage_Python)

        try:
            robot = robot_cls()
        except:
            from .driverstation import DriverStation
            DriverStation.reportError("Unhandled exception instantiating robot " + robot_cls.__name__, True)
            DriverStation.reportWarning("Robots should not quit, but yours did!", False)
            DriverStation.reportError("Could not instantiate robot "+robot_cls.__name__+"!", False)
            return False

        # Add a check to see if the user forgot to call super().__init__()
        if not hasattr(robot, '_RobotBase__initialized'):
            logger.error("If your robot class has an __init__ function, it must call super().__init__()!")
            return False

        if not hal.isSimulation():
            try:
                import wpilib
                with open('/tmp/frc_versions/FRC_Lib_Version.ini', 'w') as fp:
                    fp.write('RobotPy %s' % wpilib.__version__)
            except:
                logger.warning("Could not write FRC version file to disk")

        try:
            robot.startCompetition()
        except KeyboardInterrupt:
            logger.exception("THIS IS NOT AN ERROR: The user hit CTRL-C to kill the robot")
            logger.info("Exiting because of keyboard interrupt")
            return True
        except:
            from .driverstation import DriverStation
            DriverStation.reportError("Unhandled exception", True)
            DriverStation.reportWarning("Robots should not quit, but yours did!", False)
            DriverStation.reportError("The startCompetition() method (or methods called by it) should have handled the exception above.", False)
            return False
        else:
            # startCompetition never returns unless exception occurs....
            from .driverstation import DriverStation
            DriverStation.reportWarning("Robots should not quit, but yours did!", False)
            DriverStation.reportError("Unexpected return from startCompetition() method.", False)
            return False



# validated: 2017-12-26 EN 8b7aa61091df edu/wpi/first/wpilibj/AnalogPotentiometer.java
#----------------------------------------------------------------------------*/
# Copyright (c) FIRST 2008-2017. All Rights Reserved.                        */
# Open Source Software - may be modified and shared by FRC teams. The code   */
# must be accompanied by the FIRST BSD license file in the root directory of */
# the project.                                                               */
#----------------------------------------------------------------------------*/

import hal

from .analoginput import AnalogInput
from .interfaces import PIDSource
from .sensorbase import SensorBase

__all__ = ["AnalogPotentiometer"]

class AnalogPotentiometer(SensorBase):
    """Reads a potentiometer via an :class:`.AnalogInput`
    
    Analog potentiometers read
    in an analog voltage that corresponds to a position. The position is in
    whichever units you choose, by way of the scaling and offset constants
    passed to the constructor.

    .. not_implemented: initPot
    """
    
    PIDSourceType = PIDSource.PIDSourceType

    def __init__(self, channel, fullRange=1.0, offset=0.0):
        """AnalogPotentiometer constructor.

        Use the fullRange and offset values so that the output produces
        meaningful values. I.E: you have a 270 degree potentiometer and
        you want the output to be degrees with the halfway point as 0
        degrees. The fullRange value is 270.0(degrees) and the offset is
        -135.0 since the halfway point after scaling is 135 degrees.

        :param channel: The analog channel this potentiometer is plugged into.
        :type  channel: int or :class:`.AnalogInput`
        :param fullRange: The scaling to multiply the fraction by to get a
            meaningful unit.  Defaults to 1.0 if unspecified.
        :type  fullRange: float
        :param offset: The offset to add to the scaled value for controlling
            the zero value.  Defaults to 0.0 if unspecified.
        :type  offset: float
        """

        super().__init__()
        if not hasattr(channel, "getVoltage"):
            channel = AnalogInput(channel)
            self.addChild(channel)
        self.analog_input = channel
        self.fullRange = fullRange
        self.offset = offset
        self.init_analog_input = True
        self.pidSource = self.PIDSourceType.kDisplacement
        self.setName("AnalogPotentiometer", self.analog_input.getChannel())

    def get(self):
        """Get the current reading of the potentiometer.

        :returns: The current position of the potentiometer.
        :rtype: float
        """
        if self.analog_input is None:
            return self.offset
        return (self.analog_input.getVoltage() / hal.getUserVoltage5V()) * self.fullRange + self.offset
    
    def setPIDSourceType(self, pidSource):
        """Set which parameter you are using as a process
        control variable. 

        :param pidSource: An enum to select the parameter.
        :type  pidSource: :class:`.PIDSource.PIDSourceType`
        """
        if pidSource != self.PIDSourceType.kDisplacement:
            raise ValueError("Only displacement PID is allowed for potentiometers.")
        self.pidSource = pidSource
        
    def getPIDSourceType(self):
        return self.pidSource

    def pidGet(self):
        """Implement the PIDSource interface.

        :returns: The current reading.
        :rtype: float
        """
        return self.get()

    def initSendable(self, builder):
        if self.analog_input is not None:
            self.analog_input.initSendable(builder)

    def free(self):
        super().free()
        if self.init_analog_input:
            self.analog_input.free()
            del self.analog_input
            self.init_analog_input = False



# novalidate

import sys
import inspect

def match_arglist(name, args, kwargs, templates, allow_extra_kwargs=False):
    """
    This compares args and kwargs against the argument templates in templates.
    
    :param name: Name of the function being checked
    :param args: The list of positional arguments
    :param kwargs: The list of keyword arguments
    :param templates: A list of dictionaries corresponding to possible
    argument list formats.
    :param allow_extra_kwargs: Whether or not to allow extra keyword arguments. If this
        is true, then extra keyword arguments will be added to the result dictionary.

    An argument template is structured as follows:

    Each element in the list should be a tuple, corresponding to an argument name,
    and argument type condition. See types_match() for argument type condition structures.

    :returns The id of the selected template
    :returns A dictionary of argument name, value tuples.
    :
    """
    return __match_arglist(name, args, kwargs, templates, False, allow_extra_kwargs)
    

def _print(*args, **kwargs):
    print(*args, **kwargs)


def __match_arglist(name, args, kwargs, templates, err, allow_extra_kwargs=False):

    # TODO: we can do better at giving the user an error message...
    
    if err:
        _print("*"*50)
        _print("ERROR: Invalid arguments passed to %s()!!" % name)
        _print("       checking args against %s possible templates" % len(templates))
        _print("*"*50)
        if len(args):
            _print("Your non-keyword arguments: ")
            for i, arg in enumerate(args):
                _print("  #%d: value %s, type %s" % (i, arg, type(arg)))
        if len(kwargs):
            _print("Your keyword args: ")
            for k, v in kwargs.items():
                _print("  %s: value %s, type %s" % (k, v, type(v)))
        _print("*"*50)
    
    #Try each template until we find one that works
    for i, template in enumerate(templates):
        #List copies of the arguments
        args_copy = list(reversed(args))
        kwargs_copy = list(kwargs.copy())
        kwarg_found = False
        results = dict()
        
        if err:
            _print("Checking template %s: %s" % (i, ', '.join(an for an, _ in template)))
            showed_error = False

        #Scan through all arguments and set valid to false if we find an issue.
        for j, (arg_name, arg_type_condition) in enumerate(template):
            
            #Check kwargs first, then check args
            if arg_name in kwargs_copy:
                kwargs_copy.remove(arg_name)
                value = kwargs[arg_name]
                match_type = 'keyword'
                kwarg_found = True
            elif not kwarg_found and len(args_copy) > 0:
                value = args_copy.pop()
                match_type = 'non-keyword'
            else:
                value = None
                match_type = 'optional'

            results[arg_name] = value

            #Check to see if identities match:
            if not types_match(value, arg_type_condition):
                if err:
                    _print("- Error at arg %d: %s != %s" % (j, arg_name, typematch_to_str(arg_type_condition)))
                    _print("     your arg: %s; value %s %s" % (match_type, value, type(value)) )
                    _print()
                    showed_error = True
                break

        else:
            #If the results are valid and the argument lists are empty, return the results.
            if len(args_copy) == 0 and (len(kwargs_copy) == 0 or allow_extra_kwargs):
                output = kwargs.copy()
                output.update(results)
                return templates.index(template), output
        
        if err and not showed_error:
            if len(args_copy) != 0:
                _print("- Error: too many arguments")
            elif len(kwargs_copy):
                _print("- Error: unused parameters: %s" % ', '.join('%s' % s for s in kwargs_copy))
                
            _print()
        #    _print("Template %s unmatched:" % i)
        #    if len(args_copy):
        #        _print("- Args: %s" % (' '.join('%s' % s for s in args_copy)))
        #    if len(kwargs_copy):
        #        _print("- Kwargs: %s" % (' '.join('%s' % s for s in kwargs_copy)))

    # We found nothing, then... but we need to give the user a good
    # error message, so run it again in verbose mode, then raise the error!
    if not err:
        __match_arglist(name, args, kwargs, templates, True)
    else:
        _print("*"*50)
        raise ValueError("Attribute error, attributes given did not match any argument templates. See messages above for more info")

def types_match(object, type_structure):
    """
    :param object: the object to check
    :param type_structure: The structure, composed of lists, strings, and types,
    representing conditions that object must meet.

    Here are the possibilities for type_structure:
    - HasAttribute
    - type_condition
    - list of type_structures
    - None

    If type_structure is an HasAttribute, this will return False if
    any contained attribute is not present in object

    If type_structure is a type_condition, then object must be of the same type

    If type_structure is a list of type_structures, it is handled by running
    types_match on each item, and returning true if at least one matches.

    If type_structure is None, than everything matches it, and true is returned.
    """

    if type_structure is None:
        return True

    #Is it an attribute list?
    elif hasattr(type_structure, "matches"):
        return type_structure.matches(object)

    elif isinstance(type_structure, list) and len(type_structure) != 0:
        #If the list is not an attribute condition, check each element
        #and return true if a match is found
        for type_condition in type_structure:
            if types_match(object, type_condition):
                return True
    else:
        return isinstance(object, type_structure)

def typematch_to_str(type_structure):
    '''Only used for debugging'''
    
    if type_structure is None:
        return 'always matches'
    
    elif isinstance(type_structure, HasAttribute):
        return 'hasattr(%s)' % ' or '.join(type_structure.conditions)
    
    elif isinstance(type_structure, list):
        return ' or '.join(typematch_to_str(tc) for tc in type_structure)
    
    else:
        return '%s' % type_structure
    

class HasAttribute:
    def __init__(self, *args):
        self.conditions = args[:]

    def matches(self, object):
        for attribute in self.conditions:
            if not hasattr(object, attribute):
                return False
        
        return True


def reset_wpilib():
    '''
        Clears all devices from WPILib, and resets the hal data
    
        .. warning:: This is only intended to be used by test frameworks and
                     other debugging tools with the simulated HAL!
    '''
    
    modules = [
        'wpilib',
        'wpilib.buttons',
        'wpilib.command',
        'wpilib.interfaces'
    ]
    
    for modname in modules:
        try:
            module = sys.modules[modname]
        except KeyError:
            continue
        
        for _, cls in inspect.getmembers(module, inspect.isclass):
            if hasattr(cls, '_reset'):
                cls._reset()
                
    import hal_impl.functions
    hal_impl.functions.reset_hal()



# validated: 2018-01-19 DS 02b6615042d8 edu/wpi/first/wpilibj/RobotController.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

__all__ = ['RobotController']


class RobotController:
    """
        Contains functions for roboRIO functionality.
    """
    
    @staticmethod
    def getFPGAVersion() -> int:
        """Return the FPGA Version number. For now, expect this to be the current
        year.
        
        :returns: FPGA Version number.
        """
        return hal.getFPGAVersion()
    
    @staticmethod
    def getFPGARevision() -> int:
        """Return the FPGA Revision number. The format of the revision is 3 numbers. The 12 most
        significant bits are the Major Revision. the next 8 bits are the Minor Revision. The 12 least
        significant bits are the Build Number.
        
        :returns: FPGA Revision number.
        """
        return hal.getFPGARevision()
    
    @staticmethod
    def getFPGATime() -> int:
        """Read the microsecond timer from the FPGA.
        
        :returns: The current time in microseconds according to the FPGA.
        """
        return hal.getFPGATime()
    
    @staticmethod
    def getUserButton() -> bool:
        """Get the state of the "USER" button on the roboRIO.
        
        :returns: true if the button is currently pressed down
        """
        return hal.getFPGAButton()
    
    @staticmethod
    def getBatteryVoltage() -> float:
        """Read the battery voltage.
        
        :returns: The battery voltage in Volts.
        """
        return hal.getVinVoltage()
    
    @staticmethod
    def isSysActive() -> bool:
        """Gets a value indicating whether the FPGA outputs are enabled. The outputs may be disabled if
        the robot is disabled or e-stopped, the watchdog has expired, or if the roboRIO browns out.
        
        :returns: True if the FPGA outputs are enabled.
        """
        return hal.getSystemActive()
    
    @staticmethod
    def isBrownedOut() -> bool:
        """Check if the system is browned out.
        
        :returns: True if the system is browned out
        """
        return hal.getBrownedOut()
    
    @staticmethod
    def getInputVoltage() -> float:
        """Get the input voltage to the robot controller.
        
        :returns: The controller input voltage value in Volts
        """
        return hal.getVinVoltage()
    
    @staticmethod
    def getInputCurrent() -> float:
        """Get the input current to the robot controller.
        
        :returns: The controller input current value in Amps
        """
        return hal.getVinCurrent()
    
    @staticmethod
    def getVoltage3V3() -> float:
        """Get the voltage of the 3.3V rail.
        
        :returns: The controller 3.3V rail voltage value in Volts
        """
        return hal.getUserVoltage3V3()
    
    @staticmethod
    def getCurrent3V3() -> float:
        """Get the current output of the 3.3V rail.
        
        :returns: The controller 3.3V rail output current value in Volts
        """
        return hal.getUserCurrent3V3()
    
    @staticmethod
    def getEnabled3V3() -> bool:
        """Get the enabled state of the 3.3V rail. The rail may be disabled due to a controller brownout,
        a short circuit on the rail, or controller over-voltage.
        
        :returns: The controller 3.3V rail enabled value
        """
        return hal.getUserActive3V3()
    
    @staticmethod
    def getFaultCount3V3() -> int:
        """Get the count of the total current faults on the 3.3V rail since the controller has booted.
        
        :returns: The number of faults
        """
        return hal.getUserCurrentFaults3V3()
    
    @staticmethod
    def getVoltage5V() -> float:
        """Get the voltage of the 5V rail.
        
        :returns: The controller 5V rail voltage value in Volts
        """
        return hal.getUserVoltage5V()
    
    @staticmethod
    def getCurrent5V() -> float:
        """Get the current output of the 5V rail.
        
        :returns: The controller 5V rail output current value in Amps
        """
        return hal.getUserCurrent5V()
    
    @staticmethod
    def getEnabled5V() -> bool:
        """Get the enabled state of the 5V rail. The rail may be disabled due to a controller brownout, a
        short circuit on the rail, or controller over-voltage.
        
        :returns: The controller 5V rail enabled value
        """
        return hal.getUserActive5V()
    
    @staticmethod
    def getFaultCount5V() -> int:
        """Get the count of the total current faults on the 5V rail since the controller has booted.
        
        :returns: The number of faults
        """
        return hal.getUserCurrentFaults5V()
    
    @staticmethod
    def getVoltage6V() -> float:
        """Get the voltage of the 6V rail.
        
        :returns: The controller 6V rail voltage value in Volts
        """
        return hal.getUserVoltage6V()
    
    @staticmethod
    def getCurrent6V() -> float:
        """Get the current output of the 6V rail.
        
        :returns: The controller 6V rail output current value in Amps
        """
        return hal.getUserCurrent6V()
    
    @staticmethod
    def getEnabled6V() -> bool:
        """Get the enabled state of the 6V rail. The rail may be disabled due to a controller brownout, a
        short circuit on the rail, or controller over-voltage.
        
        :returns: The controller 6V rail enabled value
        """
        return hal.getUserActive6V()
    
    @staticmethod
    def getFaultCount6V() -> int:
        """Get the count of the total current faults on the 6V rail since the controller has booted.
        
        :returns: The number of faults
        """
        return hal.getUserCurrentFaults6V()
    
    @staticmethod
    def getCANStatus():
        """Get the current status of the CAN bus.
        
        :returns: The status of the CAN bus as a tuple: "percentBusUtilization", "busOffCount", "txFullCount", "receiveErrorCount", "transmitErrorCount"
        """
        return hal.CAN_GetCANStatus()



# validated: 2017-12-24 DS f9bece2ffbf7 edu/wpi/first/wpilibj/GyroBase.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------


from .interfaces import PIDSource
from .sensorbase import SensorBase

__all__ = ['GyroBase']

class GyroBase(SensorBase):
    '''
        GyroBase is the common base class for Gyro implementations such as
        :class:`.AnalogGyro`.
    '''
    
    PIDSourceType = PIDSource.PIDSourceType
    
    def __init__(self):
        super().__init__()
        self.pidSource = self.PIDSourceType.kDisplacement
        self.valueEntry = None
    
    def calibrate(self):
        raise NotImplementedError()
    
    def reset(self):
        raise NotImplementedError()
    
    def getAngle(self):
        raise NotImplementedError()
    
    def getRate(self):
        raise NotImplementedError()
    
    def setPIDSourceType(self, pidSource):
        """Set which parameter of the gyro you are using as a process
        control variable. The Gyro class supports the rate and angle
        parameters.

        :param pidSource: An enum to select the parameter.
        :type  pidSource: :class:`.PIDSource.PIDSourceType`
        """
        if pidSource not in (self.PIDSourceType.kDisplacement,
                             self.PIDSourceType.kRate):
            raise ValueError("Must be kRate or kDisplacement")
        self.pidSource = pidSource
        
    def getPIDSourceType(self):
        return self.pidSource
    
    def pidGet(self):
        """Get the output of the gyro for use with PIDControllers. May be
        the angle or rate depending on the set :class:`.PIDSourceType`

        :returns: the current angle according to the gyro
        :rtype: float
        """
        if self.pidSource == self.PIDSourceType.kRate:
            return self.getRate()
        elif self.pidSource == self.PIDSourceType.kDisplacement:
            return self.getAngle()
        else:
            return 0.0
    
    def initSendable(self, builder):
        builder.setSmartDashboardType("Gyro")
        builder.addDoubleProperty("Value", self.getAngle, None)





# validated: 2017-12-26 DS 7f074563d06f edu/wpi/first/wpilibj/ADXRS450_Gyro.java
#----------------------------------------------------------------------------
# Copyright (c) 2015-2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .driverstation import DriverStation
from .gyrobase import GyroBase
from .spi import SPI
from .timer import Timer

__all__ = ['ADXRS450_Gyro']

class ADXRS450_Gyro(GyroBase):
    """
    Use a rate gyro to return the robots heading relative to a starting position.
    The Gyro class tracks the robots heading based on the starting position. As
    the robot rotates the new heading is computed by integrating the rate of
    rotation returned by the sensor. When the class is instantiated, it does a
    short calibration routine where it samples the gyro while at rest to
    determine the default offset. This is subtracted from each sample to
    determine the heading.

    This class is for the digital ADXRS450 gyro sensor that connects via SPI.
    """

    kSamplePeriod = 0.0005
    kCalibrationSampleTime = 5.0
    kDegreePerSecondPerLSB = 0.0125

    kRateRegister = 0x00
    kTemRegister = 0x02
    kLoCSTRegister = 0x04
    kHiCSTRegister = 0x06
    kQuadRegister = 0x08
    kFaultRegister = 0x0A
    kPIDRegister = 0x0C
    kSNHighRegister = 0x0E
    kSNLowRegister = 0x10

    def __init__(self, port=None):
        """
            Constructor.

            :param port: The SPI port that the gyro is connected to
            :type port: :class:`.SPI.Port`
        """
        super().__init__()

        if port is None:
            port = SPI.Port.kOnboardCS0

        simPort = None
        if hal.HALIsSimulation():
            from hal_impl.spi_helpers import ADXRS450_Gyro_Sim
            simPort = ADXRS450_Gyro_Sim(self)

        self.spi = SPI(port, simPort=simPort)

        self.spi.setClockRate(3000000)
        self.spi.setMSBFirst()
        self.spi.setSampleDataOnRising()
        self.spi.setClockActiveHigh()
        self.spi.setChipSelectActiveLow()

        # Validate the part ID
        if (self._readRegister(self.kPIDRegister) & 0xff00) != 0x5200:
            self.spi.free()
            self.spi = None
            DriverStation.reportError("could not find ADXRS450 gyro on SPI port %s" % port, False)
            return

        # python-specific: make this easier to simulate
        if hal.isSimulation():
            self.spi.initAccumulator(self.kSamplePeriod, 0x20000000, 4, 0x0, 0x0, 0, 32, True, True)
        else:
            self.spi.initAccumulator(self.kSamplePeriod, 0x20000000, 4, 0x0c00000e, 0x04000000,
                10, 16, True, True)

        self.calibrate()

        hal.report(hal.UsageReporting.kResourceType_ADXRS450, port)
        self.setName("ADXRS450_Gyro", port)

    def calibrate(self):
        """Calibrate the gyro by running for a number of samples and computing the
        center value. Then use the center value as the Accumulator center value for
        subsequent measurements.

        It's important to make sure that the robot is not moving while the centering
        calculations are in progress, this is typically done when the robot is first
        turned on while it's sitting at rest before the competition starts.

        .. note:: Usually you don't need to call this, as it's called when the
                  object is first created. If you do, it will freeze your robot
                  for 5 seconds
        """

        if self.spi is None:
            return

        if not hal.HALIsSimulation():
            Timer.delay(0.1)

        self.spi.setAccumulatorCenter(0)
        self.spi.resetAccumulator()

        if not hal.HALIsSimulation():
            Timer.delay(self.kCalibrationSampleTime)

        self.spi.setAccumulatorCenter(int(self.spi.getAccumulatorAverage()))
        self.spi.resetAccumulator()

    def _calcParity(self, v):
        parity = False
        while v != 0:
            parity = not parity
            v = v & (v - 1)
        return parity

    def _readRegister(self, reg):
        cmdhi = 0x8000 | (reg << 1)
        parity = self._calcParity(cmdhi)

        data = [cmdhi >> 8,
                cmdhi & 0xff,
                0,
                0 if parity else 1]

        self.spi.write(data)
        data = self.spi.read(False, 4)

        if (data[0] & 0xe0) == 0:
            return 0  # error, return 0

        val = int.from_bytes(data[:4], byteorder='big')
        return (val >> 5) & 0xffff

    def reset(self):
        """
        Reset the gyro. Resets the gyro to a heading of zero. This can be used if
        there is significant drift in the gyro and it needs to be recalibrated
        after it has been running.
        """
        self.spi.resetAccumulator()

    def free(self):
        """Delete (free) the spi port used for the gyro and stop accumulating."""
        super().free()
        if self.spi is not None:
            self.spi.free()
            self.spi = None

    def getAngle(self):
        """
        Return the actual angle in degrees that the robot is currently facing.

        The angle is based on the current accumulator value corrected by the
        oversampling rate, the gyro type and the A/D calibration values. The angle
        is continuous, that is it will continue from 360 to 361 degrees. This
        allows algorithms that wouldn't want to see a discontinuity in the gyro
        output as it sweeps past from 360 to 0 on the second time around.

        :returns: the current heading of the robot in degrees. This heading is based
                  on integration of the returned rate from the gyro.
        """
        if self.spi is None:
            return 0.0
        return self.spi.getAccumulatorValue() * self.kDegreePerSecondPerLSB * self.kSamplePeriod

    def getRate(self):
        """Return the rate of rotation of the gyro

        The rate is based on the most recent reading of the gyro value

        :returns: the current rate in degrees per second
        """
        if self.spi is None:
            return 0.0
        else:
            return self.spi.getAccumulatorLastValue() * self.kDegreePerSecondPerLSB



# validated: 2017-12-12 EN f9bece2ffbf7 edu/wpi/first/wpilibj/Servo.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .livewindow import LiveWindow
from .pwm import PWM

__all__ = ["Servo"]

class Servo(PWM):
    """Standard hobby style servo

    The range parameters default to the appropriate values for the Hitec
    HS-322HD servo provided in the FIRST Kit of Parts in 2008.
    """

    kMaxServoAngle = 180.0
    kMinServoAngle = 0.0

    kDefaultMaxServoPWM = 2.4
    kDefaultMinServoPWM = .6

    def __init__(self, channel):
        """Constructor.

        * By default `kDefaultMaxServoPWM` ms is used as the maxPWM value
        * By default `kDefaultMinServoPWM` ms is used as the minPWM value

        :param channel: The PWM channel to which the servo is attached. 0-9 are on-board, 10-19 are on the MXP port.
        :type  channel: int
        """
        super().__init__(channel)
        self.setBounds(self.kDefaultMaxServoPWM, 0, 0, 0,
                       self.kDefaultMinServoPWM)
        self.setPeriodMultiplier(self.PeriodMultiplier.k4X)

        self.valueEntry = None
        self.valueListener = None

        hal.report(hal.UsageReporting.kResourceType_Servo,
                   self.getChannel())
        self.setName("Servo", self.getChannel())

    def set(self, value):
        """Set the servo position.

        Servo values range from 0.0 to 1.0 corresponding to the range of
        full left to full right.

        :param value: Position from 0.0 to 1.0.
        :type  value: float
        """
        self.setPosition(value)

    def get(self):
        """Get the servo position.

        Servo values range from 0.0 to 1.0 corresponding to the range of
        full left to full right.

        :returns: Position from 0.0 to 1.0.
        :rtype: float
        """
        return self.getPosition()

    def setAngle(self, degrees):
        """Set the servo angle.

        Assumes that the servo angle is linear with respect to the PWM value
        (big assumption, need to test).

        Servo angles that are out of the supported range of the servo simply
        "saturate" in that direction In other words, if the servo has a range
        of (X degrees to Y degrees) than angles of less than X result in an
        angle of X being set and angles of more than Y degrees result in an
        angle of Y being set.

        :param degrees: The angle in degrees to set the servo.
        :type  degrees: float
        """
        if degrees < self.kMinServoAngle:
            degrees = self.kMinServoAngle
        elif degrees > self.kMaxServoAngle:
            degrees = self.kMaxServoAngle

        self.setPosition(((degrees - self.kMinServoAngle)) /
                         self.getServoAngleRange())

    def getAngle(self):
        """Get the servo angle.

        Assume that the servo angle is linear with respect to the PWM value
        (big assumption, need to test).

        :returns: The angle in degrees to which the servo is set.
        :rtype: float
        """
        return self.getPosition() * self.getServoAngleRange() + self.kMinServoAngle

    def getServoAngleRange(self):
        return self.kMaxServoAngle - self.kMinServoAngle

    def initSendable(self, builder):
        builder.setSmartDashboardType("Servo")
        builder.addDoubleProperty("Value", self.get, self.set)



# validated: 2018-01-01 DV 02131639bc2c edu/wpi/first/wpilibj/DMC60.java
#----------------------------------------------------------------------------
# Copyright (c) 2008-2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

#import hal

from .pwmspeedcontroller import PWMSpeedController

__all__ = ["DMC60"]


class DMC60(PWMSpeedController):
    """
    Digilent DMC 60 Speed Controller.
    """

    def __init__(self, channel: int):
        """Constructor.

        :param channel: The PWM channel that the DMC60 is attached to.
                        0-9 are on-board, 10-19 are on the MXP port.

        .. note ::

            The DMC60 uses the following bounds for PWM values. These values
            should work reasonably well for most controllers, but if users
            experience issues such as asymmetric behavior around the deadband
            or inability to saturate the controller in either direction,
            calibration is recommended.  The calibration procedure can be
            found in the DMC 60 User Manual available from Digilent.

            - 2.004ms = full "forward"
            - 1.520ms = the "high end" of the deadband range
            - 1.500ms = center of the deadband range (off)
            - 1.480ms = the "low end" of the deadband range
            - 0.997ms = full "reverse"
        """
        super().__init__(channel)
        self.setBounds(2.004, 1.52, 1.50, 1.48, .997)
        self.setPeriodMultiplier(self.PeriodMultiplier.k1X)
        self.setSpeed(0)
        self.setZeroLatch()

        #hal.report(hal.UsageReporting.kResourceType_VictorSP, self.getChannel())
        self.setName("DMC60", self.getChannel())



# validated: 2017-12-15 EN f9bece2ffbf7 edu/wpi/first/wpilibj/Spark.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .pwmspeedcontroller import PWMSpeedController

__all__ = ["Spark"]

class Spark(PWMSpeedController):
    """
        REV Robotics SPARK Speed Controller
           
        .. not_implemented: initSpark
    """

    def __init__(self, channel):
        """Constructor.

        :param channel: The PWM channel that the SPARK is attached to. 0-9 are on-board, 10-19 are on the MXP port
        
        .. note ::
        
           Note that the SD540 uses the following bounds for PWM values. These
           values should work reasonably well for most controllers, but if users
           experience issues such as asymmetric behavior around the deadband or
           inability to saturate the controller in either direction, calibration is
           recommended. The calibration procedure can be found in the SD540 User
           Manual available from Mindsensors.
           
           - 2.003ms = full "forward"
           - 1.55ms = the "high end" of the deadband range
           - 1.50ms = center of the deadband range (off)
           - 1.46ms = the "low end" of the deadband range
           - .999ms = full "reverse"
        
        """
        super().__init__(channel)
        self.setBounds(2.003, 1.55, 1.50, 1.46, .999)
        self.setPeriodMultiplier(self.PeriodMultiplier.k1X)
        self.setSpeed(0)
        self.setZeroLatch()

        hal.report(hal.UsageReporting.kResourceType_RevSPARK,
                   self.getChannel())
        self.setName("Spark", self.getChannel())



# validated: 2018-01-01 EN f9bece2ffbf7 edu/wpi/first/wpilibj/SensorBase.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .sendablebase import SendableBase

__all__ = ["SensorBase"]

class SensorBase(SendableBase):
    """Base class for all sensors
    
    Stores most recent status information as well as containing utility
    functions for checking channels and error processing.
    """

    #: Ticks per microsecond
    kSystemClockTicksPerMicrosecond = hal.getSystemClockTicksPerMicrosecond()
    
    #: Number of digital channels per roboRIO
    kDigitalChannels = hal.getNumDigitalChannels()
    
    #: Number of analog input channels per roboRIO
    kAnalogInputChannels = hal.getNumAnalogInputs()
    
    #: Number of analog output channels per roboRIO
    kAnalogOutputChannels = hal.getNumAnalogOutputs()
    
    #: Number of solenoid channels per module
    kSolenoidChannels = hal.getNumSolenoidChannels()
    
    #: Number of PWM channels per roboRIO
    kPwmChannels = hal.getNumPWMChannels()
    
    #: Number of relay channels per roboRIO
    kRelayChannels = hal.getNumRelayHeaders()
    
    #: Number of power distribution channels per PDP
    kPDPChannels = hal.getNumPDPChannels()

    #: Number of power distribution channels per PDP
    kPDPModules = hal.getNumPDPModules()

    #: Number of PCM modules
    kPCMModules = hal.getNumPCMModules()

    #: Default solenoid module
    defaultSolenoidModule = 0

    @staticmethod
    def setDefaultSolenoidModule(moduleNumber):
        """Set the default location for the Solenoid module.

        :param moduleNumber: The number of the solenoid module to use.
        """
        SensorBase.checkSolenoidModule(moduleNumber)
        SensorBase.defaultSolenoidModule = moduleNumber

    @staticmethod
    def checkSolenoidModule(moduleNumber):
        """Verify that the solenoid module is correct.

        :param moduleNumber: The solenoid module module number to check.
        """
        if not hal.checkSolenoidModule(moduleNumber):
            raise IndexError("Requested solenoid module number %d is out of range [0, %d)." % (moduleNumber, SensorBase.kPCMModules))

    @staticmethod
    def checkDigitalChannel(channel):
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkDIOChannel(channel):
            raise IndexError("Requested digital channel number %d is out of range [0, %d)." % (channel, SensorBase.kDigitalChannels))

    @staticmethod
    def checkRelayChannel(channel):
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkRelayChannel(channel):
            raise IndexError("Requested relay channel number %d is out of range [0, %d)." % (channel, SensorBase.kRelayChannels))

    @staticmethod
    def checkPWMChannel(channel):
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkPWMChannel(channel):
            raise IndexError("Requested PWM channel number %d is out of range [0, %d)." % (channel, SensorBase.kPwmChannels))

    @staticmethod
    def checkAnalogInputChannel(channel):
        """Check that the analog input number is value.
        Verify that the analog input number is one of the legal channel
        numbers.  Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkAnalogInputChannel(channel):
            raise IndexError("Requested analog input channel number %d is out of range [0, %d)." % (channel, SensorBase.kAnalogInputChannels))

    @staticmethod
    def checkAnalogOutputChannel(channel):
        """Check that the analog input number is value.
        Verify that the analog input number is one of the legal channel
        numbers.  Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkAnalogOutputChannel(channel):
            raise IndexError("Requested analog output channel number %d is out of range [0, %d)." % (channel, SensorBase.kAnalogOutputChannels))

    @staticmethod
    def checkSolenoidChannel(channel):
        """Verify that the solenoid channel number is within limits.  Channel
        numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkSolenoidChannel(channel):
            raise IndexError("Requested solenoid channel number %d is out of range [0, %d)." % (channel, SensorBase.kSolenoidChannels))

    @staticmethod
    def checkPDPChannel(channel):
        """Verify that the power distribution channel number is within limits.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkPDPChannel(channel):
            raise IndexError("Requested PDP channel number %d is out of range [0, %d)." % (channel, SensorBase.kPDPChannels))

    @staticmethod
    def checkPDPModule(module):
        """Verify that the power distribution module number is within limits.
        Module numbers are 0-based.

        :param module: The module number to check.
        """
        if not hal.checkPDPModule(module):
            raise IndexError("Requested PDP module number %d is out of range [0, %d)." % (module, SensorBase.kPDPModules))

    @staticmethod
    def getDefaultSolenoidModule():
        """Get the number of the default solenoid module.

        :returns: The number of the default solenoid module.
        """
        return SensorBase.defaultSolenoidModule



# validated: 2017-09-27 AA e1195e8b9dab edu/wpi/first/wpilibj/SpeedController.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .pidoutput import PIDOutput

__all__ = ["SpeedController"]

class SpeedController(PIDOutput):
    """Interface for speed controlling devices."""

    def set(self, speed):
        """Common interface for setting the speed of a speed controller.

        :param speed: The speed to set.  Value should be between -1.0 and 1.0.
        """
        raise NotImplementedError

    def get(self):
        """Common interface for getting the current set speed of a speed
        controller.

        :returns: The current set speed.  Value is between -1.0 and 1.0.
        """
        raise NotImplementedError

    def setInverted(self, isInverted):
        """Common interface for inverting direction of a speed controller.

        :param isInverted: The state of inversion
        """
        raise NotImplementedError

    def getInverted(self):
        """Common interface for determining if a speed controller is in the
        inverted state or not.

        :returns: True if in inverted state
        """
        raise NotImplementedError

    def disable(self):
        """Disable the speed controller."""
        raise NotImplementedError

    def stopMotor(self):
        """Stops motor movement. Motor can be moved again by calling set without having
        to re-enable the motor.
        """
        raise NotImplementedError



# validated: 2017-12-27 TW f9bece2ffbf7 edu/wpi/first/wpilibj/BuiltInAccelerometer.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2014-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in $(WIND_BASE)/WPILib.
#----------------------------------------------------------------------------

import hal

from .interfaces import Accelerometer
from .sensorbase import SensorBase

__all__ = ["BuiltInAccelerometer"]

class BuiltInAccelerometer(SensorBase):
    """Built-in accelerometer device

    This class allows access to the roboRIO's internal accelerometer.
    """

    Range = Accelerometer.Range

    def __init__(self, range=Accelerometer.Range.k8G):
        """Constructor.
        
        :param range: The range the accelerometer will measure.  Defaults to
            +/-8g if unspecified.
        :type  range: :class:`.Accelerometer.Range`
        """
        super().__init__()
        self.setRange(range)
        self.xEntry = None
        self.yEntry = None
        self.zEntry = None
        hal.report(hal.UsageReporting.kResourceType_Accelerometer, 0, 0,
                      "Built-in accelerometer")
        self.setName("BuiltInAccel", 0)

    def setRange(self, range):
        """Set the measuring range of the accelerometer.

        :param range: The maximum acceleration, positive or negative, that
                      the accelerometer will measure.
        :type  range: :class:`BuiltInAccelerometer.Range`
        """
        
        hal.setAccelerometerActive(False)

        if range == self.Range.k2G:
            hal.setAccelerometerRange(hal.AccelerometerRange.kRange_2G)
        elif range == self.Range.k4G:
            hal.setAccelerometerRange(hal.AccelerometerRange.kRange_4G)
        elif range == self.Range.k8G:
            hal.setAccelerometerRange(hal.AccelerometerRange.kRange_8G)
        elif range == self.Range.k16G:
            raise ValueError("16G range not supported (use k2G, k4G, or k8G)")
        else:
            raise ValueError("Invalid range argument '%s'" % range)

        hal.setAccelerometerActive(True)

    def getX(self):
        """
           :returns: The acceleration of the roboRIO along the X axis in
                     g-forces
           :rtype: float
        """
        return hal.getAccelerometerX()

    def getY(self):
        """
           :returns: The acceleration of the roboRIO along the Y axis in
                     g-forces
           :rtype: float
        """
        return hal.getAccelerometerY()

    def getZ(self):
        """
           :returns: The acceleration of the roboRIO along the Z axis in
                     g-forces
           :rtype: float
        """
        return hal.getAccelerometerZ()

    def initSendable(self, builder):
        builder.setSmartDashboardType("3AxisAccelerometer")
        builder.addDoubleProperty("X", self.getX, None)
        builder.addDoubleProperty("Y", self.getY, None)
        builder.addDoubleProperty("Z", self.getZ, None)



# validated: 2017-09-24 AA e1195e8b9dab edu/wpi/first/wpilibj/PIDSource.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

__all__ = ["PIDSource"]


class PIDSource:
    """This interface allows for :class:`.PIDController` to automatically read from this
    object.
    """
    
    @staticmethod
    def from_obj_or_callable(objc):
        """
            Utility method that gets a PIDSource object
        
            :param objc: An object that implements the PIDSource interface, or
                         a callable
                         
            :returns: an object that implements the PIDSource interface 
        """
        
        # This isn't pythonic, but will cause errors to happen earlier
        if not hasattr(objc, 'pidGet'):
            if not callable(objc):
                raise TypeError("%s must be callable or ")
            return _PIDSourceWrapper(objc)
        
        if not hasattr(objc, 'getPIDSourceType'):
            raise TypeError("%s does not have getPIDSourceType method" % objc)
            
        return objc
    
        
    class PIDSourceType:
        """A description for the type of output value to provide to a
        :class:`.PIDController`"""
        kDisplacement = 0
        kRate = 1
        
    def setPIDSourceType(self, pidSource):
        """Set which parameter of the device you are using as a process control
        variable.
        
        :param pidSource: An enum to select the parameter.
        :type pidSource: :class:`.PIDSourceType`
        """
        raise NotImplementedError
        
    def getPIDSourceType(self):
        """Get which parameter of the device you are using as a process control
           variable.
           
        :returns: the currently selected PID source parameter
        """
        raise NotImplementedError

    def pidGet(self):
        """Get the result to use in :class:`.PIDController`
        
        :returns: the result to use in PIDController
        """
        raise NotImplementedError

class _PIDSourceWrapper(PIDSource):
    
    def __init__(self, fn):
        self.pidGet = fn
    
    def getPIDSourceType(self):
        return self.PIDSourceType.kDisplacement



# validated: 2018-01-27 DS e4e1eab4131d edu/wpi/first/wpilibj/command/ConditionalCommand.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2017 All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .command import Command
from .instantcommand import InstantCommand

__all__ = ['ConditionalCommand']

class ConditionalCommand(Command):
    """
        A ConditionalCommand is a :class:`.Command` that starts one of two commands.
        
        A ConditionalCommand uses m_condition to determine whether it should run m_onTrue or
        m_onFalse.
        
        A ConditionalCommand adds the proper :class:`.Command` to the :class:`.Scheduler` during
        :meth:`initialize` and then :meth:`isFinished` will
        return true once that :class:`.Command` has finished executing.
        
        If no :class:`.Command` is specified for m_onFalse, the occurrence of that condition will be a
        no-op.
        
        @see :class:`.Command`
        @see :class:`.Scheduler`
    """
    
    def __init__(self, name, onTrue=None, onFalse=None):
        """Creates a new ConditionalCommand with given name and onTrue and onFalse Commands.
        
        Users of this constructor should also override condition().
        
        :param name: the name for this command group
        :param onTrue: The Command to execute if {@link ConditionalCommand#condition()} returns true
        :param onFalse: The Command to execute if {@link ConditionalCommand#condition()} returns false
        """
        super().__init__(name)

        self.onTrue = onTrue
        self.onFalse = onFalse
        
        self.chosenCommand = None
        
        self.requireAll()

    def requireAll(self):
        if self.onTrue is not None:
            for e in self.onTrue.getRequirements():
                self.requires(e)
            
        if self.onFalse is not None:
            for e in self.onFalse.getRequirements():
                self.requires(e)
    
    def condition(self):
        """The Condition to test to determine which Command to run.
        
        :returns: true if m_onTrue should be run or false if m_onFalse should be run.
        """
        raise NotImplementedError
    
    def _initialize(self):
        '''
            Calls condition() and runs the proper command.
        '''
        if self.condition():
            self.chosenCommand = self.onTrue
        else:
            self.chosenCommand = self.onFalse
            
        if self.chosenCommand is not None:
            # This is a hack to make cancelling the chosen command inside a CommandGroup work properly
            self.chosenCommand.clearRequirements()
        
            self.chosenCommand.start()
        
        super()._initialize()
    
    def _cancel(self):
        if self.chosenCommand is not None and self.chosenCommand.isRunning():
            self.chosenCommand.cancel()
            
        super()._cancel()
    
    def isFinished(self):
        if self.chosenCommand is not None:
            return self.chosenCommand.isCompleted()
        else:
            return True
    
    def _interrupted(self):
        if self.chosenCommand is not None and self.chosenCommand.isRunning():
            self.chosenCommand.cancel()
            
        super()._interrupted()



# validated: 2018-01-19 DS 54a0a7654a91 edu/wpi/first/wpilibj/RobotDrive.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal
import math
import warnings
import weakref

from .motorsafety import MotorSafety
from .talon import Talon

__all__ = ["RobotDrive"]


def _freeRobotDrive(allocatedSpeedControllers):
    """
    Free the speed controllers if they were allocated locally
    """
    for sc in allocatedSpeedControllers:
        if hasattr(sc, "free"):
            sc.free()


class RobotDrive(MotorSafety):
    """
    .. deprecated:: 2018.0.0
        Use :class:`.DifferentialDrive` or :class:`.MecanumDrive` instead.

    Operations on a robot drivetrain based on a definition of the motor
    configuration. 

    The robot drive class handles basic driving for a robot. Currently, 2
    and 4 motor tank and mecanum drive trains are supported. In the future
    other drive types like swerve might be implemented. Motor channel numbers
    are passed supplied on creation of the class. Those are used for either
    the drive function (intended for hand created drive code, such as
    autonomous) or with the Tank/Arcade functions intended to be used for
    Operator Control driving.
    
    .. not_implemented: setupMotorSafety
    """

    class MotorType:
        """The location of a motor on the robot for the purpose of driving."""

        #: Front left
        kFrontLeft = 0

        #: Front right
        kFrontRight = 1

        #: Rear left
        kRearLeft = 2

        #: Rear right
        kRearRight = 3

    kDefaultExpirationTime = 0.1
    kDefaultSensitivity = 0.5
    kDefaultMaxOutput = 1.0

    kMaxNumberOfMotors = 4

    kArcadeRatioCurve_Reported = False
    kTank_Reported = False
    kArcadeStandard_Reported = False
    kMecanumCartesian_Reported = False
    kMecanumPolar_Reported = False

    def __init__(self, *args, **kwargs):
        """Constructor for RobotDrive.

        Either 2 or 4 motors can be passed to the constructor to implement
        a two or four wheel drive system, respectively.

        When positional arguments are used, these are the two accepted orders:
        
        - leftMotor, rightMotor
        - frontLeftMotor, rearLeftMotor, frontRightMotor, rearRightMotor

        Alternatively, the above names can be used as keyword arguments.

        Either channel numbers or motor controllers can be passed (determined
        by whether the passed object has a `set` function).  If channel
        numbers are passed, the motorController keyword argument, if present,
        is the motor controller class to use; if unspecified, :class:`.Talon` is used.
        """
        warnings.warn(
            "RobotDrive was deprecated in 2018.0.0. Use DifferentialDrive, KilloughDrive, or MecanumDrive instead.",
            DeprecationWarning, stacklevel=2)
        super().__init__()

        # keyword arguments
        self.frontLeftMotor = kwargs.pop("frontLeftMotor", None)
        self.rearLeftMotor = kwargs.pop("rearLeftMotor", None)
        self.frontRightMotor = kwargs.pop("frontRightMotor", None)
        self.rearRightMotor = kwargs.pop("rearRightMotor", None)

        if "leftMotor" in kwargs:
            self.rearLeftMotor = kwargs.pop("leftMotor")
        if "rightMotor" in kwargs:
            self.rearRightMotor = kwargs.pop("rightMotor")

        controllerClass = kwargs.pop("motorController", None)
        if controllerClass is None:
            controllerClass = Talon

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 2:
            self.rearLeftMotor = args[0]
            self.rearRightMotor = args[1]
        elif len(args) == 4:
            self.frontLeftMotor = args[0]
            self.rearLeftMotor = args[1]
            self.frontRightMotor = args[2]
            self.rearRightMotor = args[3]
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        self.allocatedSpeedControllers = list()

        # convert channel number into motor controller if needed
        if (self.frontLeftMotor is not None and
                not hasattr(self.frontLeftMotor, "set")):
            self.frontLeftMotor = controllerClass(self.frontLeftMotor)
            self.allocatedSpeedControllers.append(self.frontLeftMotor)
        if (self.rearLeftMotor is not None and
                not hasattr(self.rearLeftMotor, "set")):
            self.rearLeftMotor = controllerClass(self.rearLeftMotor)
            self.allocatedSpeedControllers.append(self.rearLeftMotor)
        if (self.frontRightMotor is not None and
                not hasattr(self.frontRightMotor, "set")):
            self.frontRightMotor = controllerClass(self.frontRightMotor)
            self.allocatedSpeedControllers.append(self.frontRightMotor)
        if (self.rearRightMotor is not None and
                not hasattr(self.rearRightMotor, "set")):
            self.rearRightMotor = controllerClass(self.rearRightMotor)
            self.allocatedSpeedControllers.append(self.rearRightMotor)

        # other defaults
        self.maxOutput = RobotDrive.kDefaultMaxOutput
        self.sensitivity = RobotDrive.kDefaultSensitivity

        # set up motor safety
        self.setExpiration(self.kDefaultExpirationTime)
        self.setSafetyEnabled(True)

        # Setup Finalizer
        self.__finalizer = weakref.finalize(self, _freeRobotDrive, self.allocatedSpeedControllers)

        # start off not moving
        self.drive(0, 0)

    def drive(self, outputMagnitude, curve):
        """Drive the motors at "outputMagnitude" and "curve".

        Both outputMagnitude and curve are -1.0 to +1.0 values, where 0.0
        represents stopped and not turning. ``curve < 0`` will turn left and ``curve > 0``
        will turn right.

        The algorithm for steering provides a constant turn radius for any normal
        speed range, both forward and backward. Increasing m_sensitivity causes
        sharper turns for fixed values of curve.

        This function will most likely be used in an autonomous routine.

        :param outputMagnitude: The speed setting for the outside wheel in a turn,
               forward or backwards, +1 to -1.
        :param curve: The rate of turn, constant for different forward speeds. Set
               ``curve < 0`` for left turn or ``curve > 0`` for right turn.

        Set ``curve = e^(-r/w)`` to get a turn radius r for wheelbase w of your robot.
        Conversely, turn radius r = -ln(curve)*w for a given value of curve and
        wheelbase w.
        """
        if not RobotDrive.kArcadeRatioCurve_Reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       self.getNumMotors(),
                       hal.UsageReporting.kRobotDrive_ArcadeRatioCurve)
            RobotDrive.kArcadeRatioCurve_Reported = True

        if curve < 0:
            value = math.log(-curve)
            ratio = (value - self.sensitivity) / (value + self.sensitivity)
            if ratio == 0:
                ratio = .0000000001
            leftOutput = outputMagnitude / ratio
            rightOutput = outputMagnitude
        elif curve > 0:
            value = math.log(curve)
            ratio = (value - self.sensitivity) / (value + self.sensitivity)
            if ratio == 0:
                ratio = .0000000001
            leftOutput = outputMagnitude
            rightOutput = outputMagnitude / ratio
        else:
            leftOutput = outputMagnitude
            rightOutput = outputMagnitude

        self.setLeftRightMotorOutputs(leftOutput, rightOutput)

    def tankDrive(self, *args, **kwargs):
        """Provide tank steering using the stored robot configuration.

        Either two joysticks (with optional specified axis) or two raw values
        may be passed positionally, along with an optional squaredInputs
        boolean.  The valid positional combinations are:

        - leftStick, rightStick
        - leftStick, rightStick, squaredInputs
        - leftStick, leftAxis, rightStick, rightAxis
        - leftStick, leftAxis, rightStick, rightAxis, squaredInputs
        - leftValue, rightValue
        - leftValue, rightValue, squaredInputs

        Alternatively, the above names can be used as keyword arguments.
        The behavior of mixes of keyword arguments in other than the
        combinations above is undefined.

        If specified positionally, the value and joystick versions are
        disambiguated by looking for a `getY` function.

        :param leftStick: The joystick to control the left side of the robot.
        :param leftAxis: The axis to select on the left side Joystick object
            (defaults to the Y axis if unspecified).
        :param rightStick: The joystick to control the right side of the robot.
        :param rightAxis: The axis to select on the right side Joystick object
            (defaults to the Y axis if unspecified).
        :param leftValue: The value to control the left side of the robot.
        :param rightValue: The value to control the right side of the robot.
        :param squaredInputs: Setting this parameter to True decreases the
            sensitivity at lower speeds.  Defaults to True if unspecified.
        """
        if not RobotDrive.kTank_Reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       self.getNumMotors(),
                       hal.UsageReporting.kRobotDrive_Tank)
            RobotDrive.kTank_Reported = True

        # keyword arguments
        leftStick = kwargs.pop("leftStick", None)
        rightStick = kwargs.pop("rightStick", None)
        leftAxis = kwargs.pop("leftAxis", None)
        rightAxis = kwargs.pop("rightAxis", None)
        leftValue = kwargs.pop("leftValue", None)
        rightValue = kwargs.pop("rightValue", None)
        squaredInputs = kwargs.pop("squaredInputs", None)

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 2 or len(args) == 3:
            left, right = args[0:2]
            if len(args) == 3:
                squaredInputs = args[2]
            # determine if stick or value
            if hasattr(left, "getY"):
                leftStick = left
            else:
                leftValue = left
            if hasattr(right, "getY"):
                rightStick = right
            else:
                rightValue = right
        elif len(args) == 4:
            leftStick, leftAxis, rightStick, rightAxis = args
        elif len(args) == 5:
            leftStick, leftAxis, rightStick, rightAxis, squaredInputs = args
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        # get value from stick if only stick provided
        if leftValue is None:
            if leftAxis is None:
                leftValue = leftStick.getY()
            else:
                leftValue = leftStick.getRawAxis(leftAxis)
        if rightValue is None:
            if rightAxis is None:
                rightValue = rightStick.getY()
            else:
                rightValue = rightStick.getRawAxis(rightAxis)

        # default to squared inputs if unspecified
        if squaredInputs is None:
            squaredInputs = True

        # square the inputs (while preserving the sign) to increase fine
        # control while permitting full power
        leftValue = RobotDrive.limit(leftValue)
        rightValue = RobotDrive.limit(rightValue)
        if squaredInputs:
            leftValue = math.copysign(leftValue * leftValue, leftValue)
            rightValue = math.copysign(rightValue * rightValue, rightValue)

        self.setLeftRightMotorOutputs(leftValue, rightValue)

    def arcadeDrive(self, *args, **kwargs):
        """Provide tank steering using the stored robot configuration.

        Either one or two joysticks (with optional specified axis) or two raw
        values may be passed positionally, along with an optional
        squaredInputs boolean.  The valid positional combinations are:

        - stick
        - stick, squaredInputs
        - moveStick, moveAxis, rotateStick, rotateAxis
        - moveStick, moveAxis, rotateStick, rotateAxis, squaredInputs
        - moveValue, rotateValue
        - moveValue, rotateValue, squaredInputs

        Alternatively, the above names can be used as keyword arguments.
        The behavior of mixes of keyword arguments in other than the
        combinations above is undefined.

        If specified positionally, the value and joystick versions are
        disambiguated by looking for a `getY` function on the stick.

        :param stick: The joystick to use for Arcade single-stick driving.
            The Y-axis will be selected for forwards/backwards and the
            X-axis will be selected for rotation rate.
        :param moveStick: The Joystick object that represents the
            forward/backward direction.
        :param moveAxis: The axis on the moveStick object to use for
            forwards/backwards (typically Y_AXIS).
        :param rotateStick: The Joystick object that represents the rotation
            value.
        :param rotateAxis: The axis on the rotation object to use for the
            rotate right/left (typically X_AXIS).
        :param moveValue: The value to use for forwards/backwards.
        :param rotateValue: The value to use for the rotate right/left.
        :param squaredInputs: Setting this parameter to True decreases the
            sensitivity at lower speeds.  Defaults to True if unspecified.
        """

        if not RobotDrive.kArcadeStandard_Reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       self.getNumMotors(),
                       hal.UsageReporting.kRobotDrive_ArcadeStandard)
            RobotDrive.kArcadeStandard_Reported = True

        # keyword arguments
        stick = kwargs.pop("stick", None)
        moveStick = kwargs.pop("moveStick", None)
        rotateStick = kwargs.pop("rotateStick", None)
        moveAxis = kwargs.pop("moveAxis", None)
        rotateAxis = kwargs.pop("rotateAxis", None)
        moveValue = kwargs.pop("moveValue", None)
        rotateValue = kwargs.pop("rotateValue", None)
        squaredInputs = kwargs.pop("squaredInputs", None)

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 1:
            stick = args[0]
        elif len(args) == 2:
            # determine if stick or value
            if hasattr(args[0], "getY"):
                stick, squaredInputs = args
            else:
                moveValue, rotateValue = args
        elif len(args) == 3:
            moveValue, rotateValue, squaredInputs = args
        elif len(args) == 4:
            moveStick, moveAxis, rotateStick, rotateAxis = args
        elif len(args) == 5:
            moveStick, moveAxis, rotateStick, rotateAxis, squaredInputs = args
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        # get value from stick if only stick provided
        if moveValue is None:
            if moveStick is None:
                moveValue = stick.getY()
            else:
                moveValue = moveStick.getRawAxis(moveAxis)
        if rotateValue is None:
            if rotateStick is None:
                rotateValue = stick.getX()
            else:
                rotateValue = rotateStick.getRawAxis(rotateAxis)

        # default to squared inputs if unspecified
        if squaredInputs is None:
            squaredInputs = True

        # local variables to hold the computed PWM values for the motors
        moveValue = RobotDrive.limit(moveValue)
        rotateValue = RobotDrive.limit(rotateValue)

        if squaredInputs:
            # square the inputs (while preserving the sign) to increase fine
            # control while permitting full power
            moveValue = math.copysign(moveValue * moveValue, moveValue)
            rotateValue = math.copysign(rotateValue * rotateValue, rotateValue)

        if moveValue > 0.0:
            if rotateValue > 0.0:
                leftMotorSpeed = moveValue - rotateValue
                rightMotorSpeed = max(moveValue, rotateValue)
            else:
                leftMotorSpeed = max(moveValue, -rotateValue)
                rightMotorSpeed = moveValue + rotateValue
        else:
            if rotateValue > 0.0:
                leftMotorSpeed = -max(-moveValue, rotateValue)
                rightMotorSpeed = moveValue + rotateValue
            else:
                leftMotorSpeed = moveValue - rotateValue
                rightMotorSpeed = -max(-moveValue, -rotateValue)

        self.setLeftRightMotorOutputs(leftMotorSpeed, rightMotorSpeed)

    def mecanumDrive_Cartesian(self, x, y, rotation, gyroAngle):
        """Drive method for Mecanum wheeled robots.

        A method for driving with Mecanum wheeled robots. There are 4 wheels
        on the robot, arranged so that the front and back wheels are toed in
        45 degrees.  When looking at the wheels from the top, the roller
        axles should form an X across the robot.

        This is designed to be directly driven by joystick axes.

        :param x: The speed that the robot should drive in the X direction.
            [-1.0..1.0]
        :param y: The speed that the robot should drive in the Y direction.
            This input is inverted to match the forward == -1.0 that
            joysticks produce. [-1.0..1.0]
        :param rotation: The rate of rotation for the robot that is
            completely independent of the translation. [-1.0..1.0]
        :param gyroAngle: The current angle reading from the gyro.  Use this
            to implement field-oriented controls.
        """
        if not RobotDrive.kMecanumCartesian_Reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       self.getNumMotors(),
                       hal.UsageReporting.kRobotDrive_MecanumCartesian)
            RobotDrive.kMecanumCartesian_Reported = True

        xIn = x
        yIn = y
        # Negate y for the joystick.
        yIn = -yIn
        # Compensate for gyro angle.
        xIn, yIn = RobotDrive.rotateVector(xIn, yIn, gyroAngle)

        wheelSpeeds = [0] * self.kMaxNumberOfMotors
        wheelSpeeds[self.MotorType.kFrontLeft] = xIn + yIn + rotation
        wheelSpeeds[self.MotorType.kFrontRight] = -xIn + yIn - rotation
        wheelSpeeds[self.MotorType.kRearLeft] = -xIn + yIn + rotation
        wheelSpeeds[self.MotorType.kRearRight] = xIn + yIn - rotation

        RobotDrive.normalize(wheelSpeeds)

        self.frontLeftMotor.set(wheelSpeeds[self.MotorType.kFrontLeft] * self.maxOutput)
        self.frontRightMotor.set(wheelSpeeds[self.MotorType.kFrontRight] * self.maxOutput)
        self.rearLeftMotor.set(wheelSpeeds[self.MotorType.kRearLeft] * self.maxOutput)
        self.rearRightMotor.set(wheelSpeeds[self.MotorType.kRearRight] * self.maxOutput)

        self.feed()

    def mecanumDrive_Polar(self, magnitude, direction, rotation):
        """Drive method for Mecanum wheeled robots.

        A method for driving with Mecanum wheeled robots. There are 4 wheels
        on the robot, arranged so that the front and back wheels are toed in
        45 degrees.  When looking at the wheels from the top, the roller
        axles should form an X across the robot.

        :param magnitude: The speed that the robot should drive in a given
            direction.
        :param direction: The direction the robot should drive in degrees.
            The direction and maginitute are independent of the rotation rate.
        :param rotation: The rate of rotation for the robot that is completely
            independent of the magnitute or direction. [-1.0..1.0]
        """
        if not RobotDrive.kMecanumPolar_Reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       self.getNumMotors(),
                       hal.UsageReporting.kRobotDrive_MecanumPolar)
            RobotDrive.kMecanumPolar_Reported = True

        # Normalized for full power along the Cartesian axes.
        magnitude = RobotDrive.limit(magnitude) * math.sqrt(2.0)
        # The rollers are at 45 degree angles.
        dirInRad = math.radians(direction + 45.0)
        cosD = math.cos(dirInRad)
        sinD = math.sin(dirInRad)

        wheelSpeeds = [0] * self.kMaxNumberOfMotors
        wheelSpeeds[self.MotorType.kFrontLeft] = sinD * magnitude + rotation
        wheelSpeeds[self.MotorType.kFrontRight] = cosD * magnitude - rotation
        wheelSpeeds[self.MotorType.kRearLeft] = cosD * magnitude + rotation
        wheelSpeeds[self.MotorType.kRearRight] = sinD * magnitude - rotation

        RobotDrive.normalize(wheelSpeeds)

        self.frontLeftMotor.set(wheelSpeeds[self.MotorType.kFrontLeft] * self.maxOutput)
        self.frontRightMotor.set(wheelSpeeds[self.MotorType.kFrontRight] * self.maxOutput)
        self.rearLeftMotor.set(wheelSpeeds[self.MotorType.kRearLeft] * self.maxOutput)
        self.rearRightMotor.set(wheelSpeeds[self.MotorType.kRearRight] * self.maxOutput)

        self.feed()

    def holonomicDrive(self, magnitude, direction, rotation):
        """Holonomic Drive method for Mecanum wheeled robots.

        This is an alias to :func:`mecanumDrive_Polar` for backward
        compatibility.

        :param magnitude: The speed that the robot should drive in a given
            direction.  [-1.0..1.0]
        :param direction: The direction the robot should drive. The direction
            and magnitude are independent of the rotation rate.
        :param rotation: The rate of rotation for the robot that is
            completely independent of the magnitude or direction.  [-1.0..1.0]
        """
        self.mecanumDrive_Polar(magnitude, direction, rotation)

    def setLeftRightMotorOutputs(self, leftOutput, rightOutput):
        """Set the speed of the right and left motors.

        This is used once an appropriate drive setup function is called such as
        twoWheelDrive(). The motors are set to "leftSpeed" and "rightSpeed"
        and includes flipping the direction of one side for opposing motors.

        :param leftOutput: The speed to send to the left side of the robot.
        :param rightOutput: The speed to send to the right side of the robot.
        """
        if self.rearLeftMotor is None or self.rearRightMotor is None:
            raise ValueError("Null motor provided")

        leftOutput = RobotDrive.limit(leftOutput) * self.maxOutput
        rightOutput = RobotDrive.limit(rightOutput) * self.maxOutput

        if self.frontLeftMotor is not None:
            self.frontLeftMotor.set(leftOutput)
        self.rearLeftMotor.set(leftOutput)

        if self.frontRightMotor is not None:
            self.frontRightMotor.set(-rightOutput)
        self.rearRightMotor.set(-rightOutput)

        self.feed()

    @staticmethod
    def limit(number):
        """Limit motor values to the -1.0 to +1.0 range."""
        if number > 1.0:
            return 1.0
        if number < -1.0:
            return -1.0
        return number

    @staticmethod
    def normalize(wheelSpeeds):
        """Normalize all wheel speeds if the magnitude of any wheel is greater
        than 1.0.
        """
        maxMagnitude = max(abs(x) for x in wheelSpeeds)
        if maxMagnitude > 1.0:
            for i in range(len(wheelSpeeds)):
                wheelSpeeds[i] = wheelSpeeds[i] / maxMagnitude

    @staticmethod
    def rotateVector(x, y, angle):
        """Rotate a vector in Cartesian space."""
        angle = math.radians(angle)
        cosA = math.cos(angle)
        sinA = math.sin(angle)
        return (x * cosA - y * sinA), (x * sinA + y * cosA)

    def setInvertedMotor(self, motor, isInverted):
        """Invert a motor direction.

        This is used when a motor should run in the opposite direction as
        the drive code would normally run it. Motors that are direct drive
        would be inverted, the drive code assumes that the motors are geared
        with one reversal.

        :param motor: The motor index to invert.
        :param isInverted: True if the motor should be inverted when operated.
        """
        if motor == self.MotorType.kFrontLeft:
            self.frontLeftMotor.setInverted(isInverted)
        elif motor == self.MotorType.kFrontRight:
            self.frontRightMotor.setInverted(isInverted)
        elif motor == self.MotorType.kRearLeft:
            self.rearLeftMotor.setInverted(isInverted)
        elif motor == self.MotorType.kRearRight:
            self.rearRightMotor.setInverted(isInverted)
        else:
            raise ValueError("Invalid motor type specified")

    def setSensitivity(self, sensitivity):
        """Set the turning sensitivity.

        This only impacts the drive() entry-point.

        :param sensitivity: Effectively sets the turning sensitivity (or turn
            radius for a given value)
        """
        self.sensitivity = sensitivity

    def setMaxOutput(self, maxOutput):
        """Configure the scaling factor for using RobotDrive with motor
        controllers in a mode other than PercentVbus.

        :param maxOutput: Multiplied with the output percentage computed by
            the drive functions.
        """
        self.maxOutput = maxOutput

    def free(self):
        self.__finalizer()
        self.frontLeftMotor = None
        self.frontRightMotor = None
        self.rearLeftMotor = None
        self.rearRightMotor = None
        self.allocatedSpeedControllers = list()
        self.setSafetyEnabled(False)

    def getDescription(self):
        return "Robot Drive"

    def stopMotor(self):
        if self.frontLeftMotor is not None:
            self.frontLeftMotor.stopMotor()
        if self.frontRightMotor is not None:
            self.frontRightMotor.stopMotor()
        if self.rearLeftMotor is not None:
            self.rearLeftMotor.stopMotor()
        if self.rearRightMotor is not None:
            self.rearRightMotor.stopMotor()
        self.feed()

    def getNumMotors(self):
        motors = 0
        if self.frontLeftMotor is not None: motors += 1
        if self.frontRightMotor is not None: motors += 1
        if self.rearLeftMotor is not None: motors += 1
        if self.rearRightMotor is not None: motors += 1
        return motors



# validated: 2017-12-07 EN f9bece2ffbf7 edu/wpi/first/wpilibj/SendableBase.java
#----------------------------------------------------------------------------
# Copyright (c) 2017 FIRST. All Rights Reserved.                             
# Open Source Software - may be modified and shared by FRC teams. The code   
# must be accompanied by the FIRST BSD license file in the root directory of 
# the project.                                                               
#----------------------------------------------------------------------------
import threading
from .sendable import Sendable
from .livewindow import LiveWindow
from ._impl.utils import match_arglist


__all__ = ['SendableBase']


class SendableBase(Sendable):
    """
    Base class for all sensors. Stores most recent status information as well as containing utility
    functions for checking channels and error processing.
    """

    def __init__(self, addLiveWindow=True):
        """
        Creates an instance of the sensor base.
   
        :param addLiveWindow: if true, add this Sendable to LiveWindow
        """
        self.name = ""
        self.subsystem = "Ungrouped"
        self.mutex = threading.RLock()

        if addLiveWindow:
            LiveWindow.add(self)

    def free(self):
        """
        Free the resources used by this object.
        """
        LiveWindow.remove(self)

    def getName(self):
        with self.mutex:
            return self.name

    def _setName(self, name):
        with self.mutex:
            self.name = name

    def setName(self, *args, **kwargs):
        """
        Sets the name of this Sendable object.

        Arguments can be structured as follows:

        - name
        - subsystem, name
        - moduleType, channel
        - moduleType, moduleNumber, channel

        :param name: name
        :type name: str
        :param subsystem: subsystem name
        :type subsystem: str
        :param moduleType: A string that defines the module name in the label for the value
        :type moduleType: str
        :param channel: The channel number the device is plugged into
        :type channel: int
        :param moduleNumber: The number of the particular module type
        :type moduleNumber: int
        """
        name_arg = ("name", [str])
        subsystem_arg = ("subsystem", [str])
        moduleType_arg = ("moduleType", [str])
        channel_arg = ("channel", [int])
        moduleNumber_arg = ("moduleNumber", [int])

        templates = [[name_arg],
                     [subsystem_arg, name_arg],
                     [moduleType_arg, channel_arg],
                     [moduleType_arg, moduleNumber_arg, channel_arg]]

        index, results = match_arglist(
            'SendableBase.setName',
            args, kwargs, templates)

        if index == 0:
            self._setName(results["name"])
        elif index == 1:
            self._setNameAndSubsystem(results["subsystem"], results["name"])
        elif index == 2:
            self._setName("{moduleType}[{channel}]".format_map(results))
        elif index == 3:
            self._setName("{moduleType}[{moduleNumber},{channel}]".format_map(results))

    def getSubsystem(self):
        with self.mutex:
            return self.subsystem

    def setSubsystem(self, subsystem):
        with self.mutex:
            self.subsystem = subsystem

    def addChild(self, child):
        """
        Add a child component

        :param child: child component
        """
        LiveWindow.addChild(self, child)




# validated: 2017-12-27 TW f9bece2ffbf7 edu/wpi/first/wpilibj/ADXL362.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .driverstation import DriverStation
from .interfaces import Accelerometer
from .spi import SPI
from .sensorbase import SensorBase

__all__ = ["ADXL362"]

class ADXL362(SensorBase):
    """
        ADXL362 SPI Accelerometer.
    
        This class allows access to an Analog Devices ADXL362 3-axis accelerometer.
        
        .. not_implemented: init
    """
    
    kRegWrite = 0x0A
    kRegRead = 0x0B
     
    kPartIdRegister = 0x02
    kDataRegister = 0x0E
    kFilterCtlRegister = 0x2C
    kPowerCtlRegister = 0x2D
    
    kFilterCtl_Range2G = 0x00
    kFilterCtl_Range4G = 0x40
    kFilterCtl_Range8G = 0x80
    kFilterCtl_ODR_100Hz = 0x03
    
    kPowerCtl_UltraLowNoise = 0x20
    kPowerCtl_AutoSleep = 0x04
    kPowerCtl_Measure = 0x02

    Range = Accelerometer.Range

    class Axes:
        kX = 0x00
        kY = 0x02
        kZ = 0x04

    def __init__(self, range, port=None):
        """Constructor.
        
        :param range: The range (+ or -) that the accelerometer will measure.
        :type range: :class:`.ADXL362.Range`
        :param port: The SPI port that the accelerometer is connected to
        :type port: :class:`.SPI.Port`
        """
        if port is None:
            port = SPI.Port.kOnboardCS1
        
        self.spi = SPI(port)
        self.spi.setClockRate(3000000)
        self.spi.setMSBFirst()
        self.spi.setSampleDataOnFalling()
        self.spi.setClockActiveLow()
        self.spi.setChipSelectActiveLow()

        # Validate the part ID
        data = [self.kRegRead, self.kPartIdRegister, 0]
        data = self.spi.transaction(data)
        if data[2] != 0xF2:
            DriverStation.reportError("could not find ADXL362 on SPI port " + port, False)
            self.spi.free()
            self.spi = None
            return
    
        self.setRange(range)

        # Turn on the measurements
        self.spi.write([self.kRegWrite, self.kPowerCtlRegister,
                        self.kPowerCtl_Measure | self.kPowerCtl_UltraLowNoise])

        hal.report(hal.UsageReporting.kResourceType_ADXL362,
                      port)

        self.setName("ADXL362", port)

    def free(self):
        if self.spi:
            self.spi.free()
            self.spi = None
        super().free()

    # Accelerometer interface

    def setRange(self, range):
        """Set the measuring range of the accelerometer.

        :param range: The maximum acceleration, positive or negative, that
                      the accelerometer will measure.
        :type  range: :class:`ADXL362.Range`
        """
        if not self.spi:
            return

        if range == self.Range.k2G:
            value = self.kFilterCtl_Range2G
            self.gsPerLSB = 0.001
        elif range == self.Range.k4G:
            value = self.kFilterCtl_Range4G
            self.gsPerLSB = 0.002
        # 16G not supported; treat as 8G
        elif range == self.Range.k8G or \
             range == self.Range.k16G:
            value = self.kFilterCtl_Range8G
            self.gsPerLSB = 0x004
        else:
            raise ValueError("Invalid range argument '%s'" % range)

        self.spi.write([self.kRegWrite, self.kFilterCtlRegister,
                        self.kFilterCtl_ODR_100Hz | value])

    def getX(self):
        """Get the x axis acceleration

        :returns: The acceleration along the x axis in g-forces
        """
        return self.getAcceleration(self.Axes.kX)

    def getY(self):
        """Get the y axis acceleration

        :returns: The acceleration along the y axis in g-forces
        """
        return self.getAcceleration(self.Axes.kY)

    def getZ(self):
        """Get the z axis acceleration

        :returns: The acceleration along the z axis in g-forces
        """
        return self.getAcceleration(self.Axes.kZ)

    def getAcceleration(self, axis):
        """Get the acceleration of one axis in Gs.

        :param axis: The axis to read from.
        :returns: An object containing the acceleration measured on each axis in Gs.
        """
        if self.spi is None:
            return 0.0
        
        data = [self.kRegRead,
                self.kDataRegister + axis, 0, 0]
        data = self.spi.transaction(data)
        # Sensor is little endian... swap bytes
        rawAccel = (data[2] << 8) | data[1]
        return rawAccel * self.gsPerLSB

    def getAccelerations(self):
        """Get the acceleration of all axes in Gs.

        :returns: X,Y,Z tuple of acceleration measured on all axes in Gs.
        """
        if self.spi is None:
            return 0.0, 0.0, 0.0
        
        # Select the data address.
        data = [0] * 8
        data[0] = self.kRegRead
        data[1] = self.kDataRegister
        data = self.spi.transaction(data)

        # Sensor is little endian... swap bytes
        rawData = []
        for i in range(3):
            rawData.append((data[i*2+2] << 8) | data[i*2+1])

        return (rawData[0] * self.gsPerLSB,
                rawData[1] * self.gsPerLSB,
                rawData[2] * self.gsPerLSB)

    def _updateValues(self):
        data = self.getAccelerations()
        self._entryX.setDouble(data[0])
        self._entryY.setDouble(data[1])
        self._entryZ.setDouble(data[2])

    def initSendable(self, builder):
        builder.setSmartDashboardType("3AxisAccelerometer")
        self._entryX = builder.getEntry("X")
        self._entryY = builder.getEntry("Y")
        self._entryZ = builder.getEntry("Z")

        builder.setUpdateTable(self._updateValues)



# validated: 2018-01-06 DS 479d0beb5a79 edu/wpi/first/wpilibj/SerialPort.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import warnings
import weakref

from .resource import Resource

__all__ = ['SerialPort']

import logging
logger = logging.getLogger(__name__)

def _freeSerialPort(port):
    hal.closeSerial(port)

class SerialPort:
    """
        Driver for the RS-232 serial port on the roboRIO.
        
        The current implementation uses the VISA formatted I/O mode. This means that all traffic goes
        through the formatted buffers. This allows the intermingled use of print(), readString(), and the
        raw buffer accessors read() and write().
        
        More information can be found in the NI-VISA User Manual here:
        http://www.ni.com/pdf/manuals/370423a.pdf
        
        and the NI-VISA Programmer's Reference Manual here:
        http://www.ni.com/pdf/manuals/370132c.pdf
    """
    
    class Port:
        kOnboard = 0
        kMXP = 1
        kUSB = 2
        kUSB1 = 2
        kUSB2 = 3
            
    class Parity:
        kNone = 0
        kOdd = 1
        kEven = 2
        kMark = 3
        kSpace = 4
            
    class StopBits:
        kOne = 10
        kOnePointFive = 15
        kTwo = 20
            
    class FlowControl:
        kNone = 0
        kXonXoff = 1
        kRtsCts = 2
        kDtsDsr = 4
            
    class WriteBufferMode:
        kFlushOnAccess = 1
        kFlushWhenFull = 2
    
    
    def __init__(self, baudRate, port, dataBits=8,
                 parity=Parity.kNone,
                 stopBits=StopBits.kOne,
                 simPort=None):
        """Create an instance of a Serial Port class.
        
        :param baudRate: The baud rate to configure the serial port.
        :param port: The Serial port to use
        :param dataBits: The number of data bits per transfer. Valid values are between 5 and 8 bits.
        :param parity: Select the type of parity checking to use.
        :param stopBits: The number of stop bits to use as defined by the enum StopBits.
        :param simPort: This must be an object that implements all of
                        the serial* functions from hal_impl that you use.
                        See ``test_serial.py`` for an example.
        """
    
        if port not in [self.Port.kOnboard,
                        self.Port.kMXP,
                        self.Port.kUSB,
                        self.Port.kUSB1,
                        self.Port.kUSB2]:
            raise ValueError("Invalid value '%s' for serial port" % port)
        
        if hal.isSimulation():
            if simPort is None:
                # If you want more functionality, implement your own mock
                from hal_impl.serial_helpers import SerialSimBase
                simPort = SerialSimBase()
                
                msg = "Using stub simulator for Serial port %s" % port
                warnings.warn(msg)
        
            self._port = (simPort, port)
        else:
            self._port = port
        
        hal.initializeSerialPort(self._port)
        self.__finalizer = weakref.finalize(self, _freeSerialPort, self._port)
        
        hal.setSerialBaudRate(self.port, baudRate)
        hal.setSerialDataBits(self.port, dataBits)
        hal.setSerialParity(self.port, parity)
        hal.setSerialStopBits(self.port, stopBits)

        # Set the default read buffer size to 1 to return bytes immediately
        self.setReadBufferSize(1)

        # Set the default timeout to 5 seconds.
        self.setTimeout(5.0)

        # Don't wait until the buffer is full to transmit.
        self.setWriteBufferMode(self.WriteBufferMode.kFlushOnAccess)

        self.disableTermination()

        hal.report(hal.UsageReporting.kResourceType_SerialPort, 0)
        
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

    @property
    def port(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use serial port after free() has been called")
        return self._port
    
    def free(self):
        '''Destructor'''
        self._finalizer()

    def setFlowControl(self, flowControl):
        """Set the type of flow control to enable on this port.
        
        By default, flow control is disabled.
        
        :param flowControl: the FlowControl value to use
        """
        hal.setSerialFlowControl(self.port, flowControl)
    
    def enableTermination(self, terminator=b'\n'):
        """Enable termination and specify the termination character.
        
        Termination is currently only implemented for receive. When the the terminator is received,
        the :meth:`read` or :meth:`readString` will return fewer bytes than requested, stopping after the
        terminator.
        
        :param terminator: The character to use for termination (default is ``\\n``).
        """
        hal.enableSerialTermination(self.port, terminator)
    
    def disableTermination(self):
        """Disable termination behavior."""
        hal.disableSerialTermination(self.port)
    
    def getBytesReceived(self):
        """Get the number of bytes currently available to read from the serial port.
        
        :returns: The number of bytes available to read.
        """
        return hal.getSerialBytesReceived(self.port)
    
    def readString(self, count=None):
        """Read a string out of the buffer. Reads the entire contents of the buffer
        
        :param count: the number of characters to read into the string
        :returns: The read string
        """
        if count is None:
            count = self.getBytesReceived()
        
        try:
            return self.read(count).decode('ascii')
        except UnicodeDecodeError:
            logger.warning("Error decoding serial port output")
            return ''
    
    def read(self, count):
        """Read raw bytes out of the buffer.
        
        :param count: The maximum number of bytes to read.
        :returns: A list containing the read bytes
        """
        return hal.readSerial(self.port, count)
    
    def write(self, buffer):
        """Write raw bytes to the serial port.
        
        :param buffer: The buffer of bytes to write.
        :param count: The maximum number of bytes to write.
        :returns: The number of bytes actually written into the port.
        """
        return hal.writeSerial(self.port, buffer)
    
    def writeString(self, data):
        """Write an ASCII encoded string to the serial port
        
        :param data: The string to write to the serial port.
        :returns: The number of bytes actually written into the port.
        """
        return self.write(data.encode('ascii'))
    
    def setTimeout(self, timeout):
        """Configure the timeout of the serial self.port.
        
        This defines the timeout for transactions with the hardware. It will affect reads if less
        bytes are available than the read buffer size (defaults to 1) and very large writes.
        
        :param timeout: The number of seconds to to wait for I/O.
        """
        hal.setSerialTimeout(self.port, timeout)
    
    def setReadBufferSize(self, size):
        """Specify the size of the input buffer.
        
        Specify the amount of data that can be stored before data from the device is returned to
        Read. If you want data that is received to be returned immediately, set this to 1.
        
        It the buffer is not filled before the read timeout expires, all data that has been received
        so far will be returned.
        
        :param size: The read buffer size.
        """
        hal.setSerialReadBufferSize(self.port, size)
    
    def setWriteBufferSize(self, size):
        """Specify the size of the output buffer.
        
        Specify the amount of data that can be stored before being transmitted to the device.
        
        :param size: The write buffer size.
        """
        hal.setSerialWriteBufferSize(self.port, size)
    
    def setWriteBufferMode(self, mode):
        """Specify the flushing behavior of the output buffer.
        
        When set to kFlushOnAccess, data is synchronously written to the serial port after each
        call to either print() or write().
        
        When set to kFlushWhenFull, data will only be written to the serial port when the buffer
        is full or when flush() is called.
        
        :param mode: The write buffer mode.
        :type mode: WriteBufferMode
        """
        hal.setSerialWriteMode(self.port, mode)
    
    def flush(self):
        """Force the output buffer to be written to the port.
        
        This is used when :meth:`setWriteBufferMode` is set to kFlushWhenFull to force a flush before the
        buffer is full.
        """
        hal.flushSerial(self.port)
    
    def reset(self):
        """Reset the serial port driver to a known state.
        
        Empty the transmit and receive buffers in the device and formatted I/O.
        """
        hal.clearSerial(self.port)




# validated: 2017-11-13 TW 21585f70a88e edu/wpi/first/wpilibj/GenericHID.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import enum

from ..driverstation import DriverStation
import hal

__all__ = ["GenericHID"]


class GenericHID:
    """
    GenericHID Interface.
    """

    class RumbleType(enum.IntEnum):
        """Represents a rumble output on the JoyStick."""

        #: Left Hand
        kLeftRumble = 0

        #: Right Hand
        kRightRumble = 1

    class HIDType(enum.IntEnum):
        kUnknown = -1
        kXInputUnknown = 0
        kXInputGamepad = 1
        kXInputWheel = 2
        kXInputArcadeStick = 3
        kXInputFlightStick = 4
        kXInputDancePad = 5
        kXInputGuitar = 6
        kXInputGuitar2 = 7
        kXInputDrumKit = 8
        kXInputGuitar3 = 11
        kXInputArcadePad = 19
        kHIDJoystick = 20
        kHIDGamepad = 21
        kHIDDriving = 22
        kHIDFlight = 23
        kHID1stPerson = 24

    class Hand(enum.IntEnum):
        """Which hand the Human Interface Device is associated with."""

        #: Left Hand
        kLeft = 0

        #: Right Hand
        kRight = 1

    def __init__(self, port: int) -> None:
        self.port = port
        self.ds = DriverStation.getInstance()
        self.outputs = 0
        self.leftRumble = 0
        self.rightRumble = 0

    def getX(self, hand: Hand = Hand.kRight) -> float:
        """Get the x position of HID.

        :param hand: which hand, left or right
        :returns: the x position
        """
        raise NotImplementedError

    def getY(self, hand: Hand = Hand.kRight) -> float:
        """Get the y position of the HID.

        :param hand: which hand, left or right
        :returns: the y position
        """
        raise NotImplementedError

    def getRawButton(self, button: int) -> bool:
        """Get the button value (starting at button 1).

        :param button: The button number to be read (starting at 1)
        :returns: The state of the button.
        """
        return self.ds.getStickButton(self.port, button)

    def getRawButtonPressed(self, button: int) -> bool:
        """Whether the button was pressed since the last check. Button indexes begin at 1.

        :param button: The button index, beginning at 1.
        :returns: Whether the button was pressed since the last check.

        .. versionadded:: 2018.0.0
        """
        return self.ds.getStickButtonPressed(self.port, button)

    def getRawButtonReleased(self, button: int) -> bool:
        """Whether the button was released since the last check. Button indexes begin at 1.

        :param button: The button index, beginning at 1.
        :returns: Whether the button was released since the last check.

        .. versionadded:: 2018.0.0
        """
        return self.ds.getStickButtonReleased(self.port, button)

    def getRawAxis(self, axis: int) -> float:
        """Get the raw axis.

        :param axis: index of the axis
        :returns: the raw value of the selected axis
        """
        return self.ds.getStickAxis(self.port, axis)

    def getPOV(self, pov: int = 0) -> int:
        """Get the angle in degrees of a POV on the HID.

        The POV angles start at 0 in the up direction, and increase clockwise (eg right is 90,
        upper-left is 315).

        :param pov: The index of the POV to read (starting at 0)
        :returns: the angle of the POV in degrees, or -1 if the POV is not pressed.
        """
        return self.ds.getStickPOV(self.port, pov)

    def getAxisCount(self) -> int:
        """Get the number of axes for the HID

        :returns: The number of axis for the current HID
        """
        return self.ds.getStickAxisCount(self.port)

    def getPOVCount(self) -> int:
        """For the current HID, return the number of POVs."""
        return self.ds.getStickPOVCount(self.port)

    def getButtonCount(self) -> int:
        """For the current HID, return the number of buttons."""
        return self.ds.getStickButtonCount(self.port)

    def getPort(self) -> int:
        """Get the port number of the HID.

        :returns: The port number of the HID.
        """
        return self.port

    def getType(self) -> HIDType:
        """Get the type of the HID.

        :returns: the type of the HID.
        """
        return self.HIDType(self.ds.getJoystickType(self.port))

    def getName(self) -> str:
        """Get the name of the HID.

        :returns: the name of the HID.
        """
        return self.ds.getJoystickName(self.port)

    def setOutput(self, outputNumber: int, value: bool) -> None:
        """Set a single HID output value for the HID.

        :param outputNumber: The index of the output to set (1-32)
        :param value: The value to set the output to
        """
        self.outputs = (self.outputs & ~(1 << (outputNumber - 1))) | ((1 if value else 0) << (outputNumber - 1))
        hal.setJoystickOutputs(self.port, self.outputs, self.leftRumble, self.rightRumble)

    def setOutputs(self, value: int) -> None:
        """Set all HID output values for the HID.

        :param value: The 32 bit output value (1 bit for each output)
        """
        self.outputs = value
        hal.setJoystickOutputs(self.port, self.outputs, self.leftRumble, self.rightRumble)

    def setRumble(self, type: RumbleType, value: float) -> None:
        """Set the rumble output for the HID. The DS currently supports 2 rumble values, left rumble and
        right rumble.

        :param type: Which rumble value to set
        :param value: The normalized value (0 to 1) to set the rumble to
        """
        if value < 0:
            value = 0
        elif value > 1:
            value = 1

        if type == self.RumbleType.kLeftRumble:
            self.leftRumble = int(value * 65535)
        else:
            self.rightRumble = int(value * 65535)

        hal.setJoystickOutputs(self.port, self.outputs, self.leftRumble, self.rightRumble)
